from collections import defaultdict

import csv

import json

import logging

from multiqc.modules.base_module import BaseMultiqcModule
from multiqc.plots import bargraph, box, table

log = logging.getLogger(__name__)


REFINE_CATEGORIES = {
    "fivelen": "5' primer length",
    "threelen": "3' primer length",
    "insertlen": "Insert length",
    "polyAlen": "Poly(A) length",
}


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        super(MultiqcModule, self).__init__(
            name="Iso-Seq",
            anchor="isoseq",
            href="https://github.com/PacificBiosciences/IsoSeq",
            info="contains the newest tools to identify transcripts in PacBio single-molecule sequencing data (HiFi reads).",
            # doi=,  # Not published
        )

        cnt_by_cluster_id_by_sample = self._parse_cluster()
        refine_json_data_by_sample, refine_csv_data_by_sample = self._parse_refine()

        if not cnt_by_cluster_id_by_sample and not refine_json_data_by_sample and not refine_csv_data_by_sample:
            raise UserWarning

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        if cnt_by_cluster_id_by_sample:
            self._add_general_stats_cluster(cnt_by_cluster_id_by_sample)
            self._add_cluster_size_plot(cnt_by_cluster_id_by_sample)

        if refine_json_data_by_sample:
            self._add_refine_table(refine_json_data_by_sample)
            self._add_refine_box_plots(refine_csv_data_by_sample)

    def _parse_refine(self):
        refine_json_data_by_sample = dict()
        refine_csv_data_by_sample = dict()
        for f in self.find_log_files("isoseq/refine-json", filehandles=True):
            refine_json_data_by_sample[f["s_name"]] = json.load(f["f"])
            self.add_data_source(f, section="refine-json")

        for f in self.find_log_files("isoseq/refine-csv", filehandles=True):
            reader: csv.DictReader = csv.DictReader(f["f"])
            file_cols = reader.fieldnames
            expected_cols = set(REFINE_CATEGORIES.keys())
            missing_cols = set(expected_cols) - set(file_cols)
            if missing_cols:
                log.warning(f"Expected columns {missing_cols} not found in {f['fn']}")

            found_cols = set(file_cols) & set(expected_cols)
            vals_by_metric = defaultdict(list)
            for row in reader:
                for col in found_cols:
                    vals_by_metric[col].append(int(row[col]))
            refine_csv_data_by_sample[f["s_name"]] = vals_by_metric
            self.add_data_source(f, section="refine-csv")

        refine_json_data_by_sample = self.ignore_samples(refine_json_data_by_sample)
        refine_csv_data_by_sample = self.ignore_samples(refine_csv_data_by_sample)
        log.info(f"Found {len(refine_json_data_by_sample)} refine reports")
        log.info(f"Found {len(refine_csv_data_by_sample)} refine reports")
        if refine_json_data_by_sample:
            self.write_data_file(refine_json_data_by_sample, "multiqc_isoseq_refine_csv")
        if refine_csv_data_by_sample:
            self.write_data_file(refine_csv_data_by_sample, "multiqc_isoseq_refine_json")

        return refine_json_data_by_sample, refine_csv_data_by_sample

    def _parse_cluster(self):
        cnt_by_cluster_id_by_sample = dict()
        for f in self.find_log_files("isoseq/cluster-csv", filehandles=True):
            cnt_by_cluster_id = defaultdict(int)
            reader: csv.DictReader = csv.DictReader(f["f"])
            for row in reader:
                cluster_id = row.get("cluster_id", None)
                if cluster_id is not None:
                    cnt_by_cluster_id[cluster_id] += 1
            if cnt_by_cluster_id:
                cnt_by_cluster_id_by_sample[f["s_name"]] = cnt_by_cluster_id
                self.add_data_source(f)
        if cnt_by_cluster_id_by_sample:
            self.write_data_file(cnt_by_cluster_id_by_sample, "multiqc_isoseq_cluster")
        cnt_by_cluster_id_by_sample = self.ignore_samples(cnt_by_cluster_id_by_sample)
        log.info(f"Found {len(cnt_by_cluster_id_by_sample)} cluster reports")
        if cnt_by_cluster_id_by_sample:
            self.write_data_file(cnt_by_cluster_id_by_sample, "multiqc_isoseq_cluster")
        return cnt_by_cluster_id_by_sample

    def _add_general_stats_cluster(self, size_by_cluster_id_by_sample):
        gstats_data = {}
        for s_name, size_by_cluster_id in size_by_cluster_id_by_sample.items():
            gstats_data[s_name] = {}
            gstats_data[s_name]["n_cluster"] = len(size_by_cluster_id)
            gstats_data[s_name]["mean_cluster_size"] = sum(size_by_cluster_id.values()) / len(size_by_cluster_id)

        headers = {
            "n_cluster": {
                "title": "Clusters",
                "description": "Number of clusters created during the clustering. (1 cluster = 1 transcript)",
                "format": "{:,.d}",
                "scale": "Spectral",
            },
            "mean_cluster_size": {
                "title": "Mean cluster size",
                "description": "Average number of CCS clustered to form transcripts.",
                "format": "{:,.2f}",
                "scale": "RdYlGn",
            },
        }

        self.general_stats_addcols(gstats_data, headers)

    def _add_cluster_size_plot(self, cnt_by_cluster_id_by_sample):
        plot_data = dict()

        for s_name, size_by_cluster_id in cnt_by_cluster_id_by_sample.items():
            plot_data[s_name] = {"=2": 0, "3-10": 0, "11-100": 0, ">100": 0}
            value_counts = defaultdict(int)

            # Calculating value counts similar to df.value_counts("n_CCS") in pandas
            for key, size in size_by_cluster_id.items():
                value_counts[size] += 1

            for n_CCS, size in value_counts.items():
                if n_CCS == 2:
                    plot_data[s_name]["=2"] = size
                elif 2 < n_CCS < 11:
                    plot_data[s_name]["3-10"] += size
                elif 10 < n_CCS < 101:
                    plot_data[s_name]["11-100"] += size
                elif n_CCS > 100:
                    plot_data[s_name][">100"] += size

        cats = {
            "=2": {
                "name": "2 CCS",
            },
            "3-10": {
                "name": "3 to 10 CCS",
            },
            "11-100": {
                "name": "11 to 100 CCS",
            },
            ">100": {
                "name": "More than 100 CCS",
            },
        }

        self.add_section(
            name="Cluster size distribution",
            anchor="isoseq-cluster-size-distribution",
            description="A distribution of cluster size (number of CC clustered to form one Hifi read)",
            helptext="""
            The CSV report <code>*.cluster_report.csv</code> files produced by Iso-Seq <code>cluster</code> 
            shows which CCS have been clustered together to form one Hifi reads. The bargraph represent the 
            distribution of the cluster size using four categories : -2, 3-10, 11-100, >100.
            """,
            plot=bargraph.plot(
                plot_data,
                cats,
                {
                    "id": "isoseq-cluster-size-distribution-barplot",
                    "title": "Iso-Seq cluster: Histogram of cluster size",
                    "ylab": "Count",  # Y axis label
                },
            ),
        )

    def _add_refine_table(self, data_by_sample):
        headers = {
            "num_reads_fl": {
                "title": "Full-length",
                "description": "Number of CCS where both primers have been detected",
                "scale": "GnBu",
                "format": "{:,.d}",
            },
            "num_reads_flnc": {
                "title": "Non-chimeric full-length",
                "description": "Number of non-chimeric CCS where both primers have been detected",
                "scale": "RdYlGn",
                "format": "{:,.d}",
            },
            "num_reads_flnc_polya": {
                "title": "Poly(A) free non-chimeric full-length",
                "description": (
                    "Number of non-chimeric CCS where both primers have been detected and the poly(A) "
                    "tail has been removed"
                ),
                "scale": "GnBu",
                "format": "{:,.d}",
            },
        }
        return self.add_section(
            name="CCS statistics",
            anchor="insert-refine-stats",
            description="Statistics from the Iso-Seq <code>refine</code> summary JSON files.",
            helptext="""
            Iso-Seq <code>refine</code> performs the full-length read refinement by trimming poly(A) tails and
            removing concatemers. It also outputs a summary JSON file <code>*.filter_summary.json</code> per each 
            CCS file with statistics of refined full-length non-concatemer reads.
            """,
            plot=table.plot(
                data_by_sample,
                headers,
                {
                    "id": "isoseq_refine_bargraph",
                    "title": "Iso-Seq refine: filter summary",
                },
            ),
        )

    def _add_refine_box_plots(self, values_by_metric_by_sample):
        data_by_sample_by_metric = {k: {} for k in REFINE_CATEGORIES.keys()}
        for sname, values_by_metric in values_by_metric_by_sample.items():
            for metric in REFINE_CATEGORIES.keys():
                data_by_sample_by_metric[metric][sname] = values_by_metric[metric]

        for metric, data_by_sample in data_by_sample_by_metric.items():
            self.add_section(
                name=f"{REFINE_CATEGORIES[metric]}",
                anchor=f"insert-refine-stats-{metric}",
                helptext="""
                Iso-Seq <code>refine</code> performs the full-length read refinement by trimming poly(A) tails and
                removing concatemers. It also outputs a report table <code>*.report.csv</code> per each CCS file,
                providing statistics on each CCS: 5' prime and 3' primer lengths, insert length, poly(A) length, 
                and couple of primers detected for each CCS. The box plots present the distribution showing the 
                min, max, mean, standard deviation statistics for each metric.
                """,
                plot=box.plot(
                    list_of_data_by_sample=data_by_sample,
                    pconfig={
                        "id": f"isoseq_refine_boxplot_{metric}",
                        "title": f"Iso-Seq refine: {REFINE_CATEGORIES[metric]}",
                    },
                ),
            )



""" MultiQC module to parse output from Cluster Flow """


import datetime
import logging
import os
import re
import time

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import table

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """
    Cluster Flow module class, parses run logs.
    """

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Cluster Flow",
            anchor="clusterflow",
            href="http://clusterflow.io",
            info="is a simple and flexible bioinformatics pipeline tool.",
            doi="10.12688/f1000research.10335.2",
        )

        # Find and load any Cluster Flow reports
        self.clusterflow_commands = dict()
        self.clusterflow_runfiles = dict()

        for f in self.find_log_files("clusterflow/logs", filehandles=True):
            self.parse_clusterflow_logs(f)
            self.add_data_source(f, "log")

        for f in self.find_log_files("clusterflow/runfiles", filehandles=True):
            parsed_data = self.parse_clusterflow_runfiles(f)
            if parsed_data is not None:
                self.clusterflow_runfiles[f["s_name"]] = parsed_data
                self.add_data_source(f, "runfile")

        # Filters to strip out ignored sample names
        self.clusterflow_commands = self.ignore_samples(self.clusterflow_commands)
        self.clusterflow_runfiles = self.ignore_samples(self.clusterflow_runfiles)

        if len(self.clusterflow_commands) == 0 and len(self.clusterflow_runfiles) == 0:
            raise ModuleNoSamplesFound

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Count pipelines
        num_log_pipelines = len(self.clusterflow_commands)
        num_runfile_pipelines = len(set(d.get("pipeline_id", "unknown") for d in self.clusterflow_runfiles.values()))
        log.info(f"Found {max(num_log_pipelines, num_runfile_pipelines)} pipelines")
        log.debug(f"Found {num_log_pipelines} log pipelines")
        log.debug(f"Found {num_runfile_pipelines} runfile pipelines")

        # Pipeline Info
        if len(self.clusterflow_runfiles) > 0:
            self.clusterflow_pipelines_section()

            # Write data to file
            self.write_data_file(self.clusterflow_runfiles, "clusterflow_runfiles")

        # Commands
        if len(self.clusterflow_commands) > 0:
            self.clusterflow_commands_table()

            # Write data to file
            self.write_data_file(self.clusterflow_runfiles, "clusterflow_commands")

    def parse_clusterflow_logs(self, f):
        """Parse Clusterflow logs"""
        module = None
        job_id = None
        pipeline_id = None
        for line in f["f"]:
            # Get pipeline ID
            module_r = re.match(r"Module:\s+(.+)$", line)
            if module_r:
                module = module_r.group(1)
            job_id_r = re.match(r"Job ID:\s+(.+)$", line)
            if job_id_r:
                job_id = job_id_r.group(1)
                if module is not None:
                    pipeline_r = re.match(r"(cf_.+)_" + re.escape(module) + r"_\d+$", job_id)
                    if pipeline_r:
                        pipeline_id = pipeline_r.group(1)

            # Get commands that have been run
            if line.startswith("###CFCMD"):
                if pipeline_id is None:
                    pipeline_id = "unknown"
                if pipeline_id not in self.clusterflow_commands.keys():
                    self.clusterflow_commands[pipeline_id] = list()
                self.clusterflow_commands[pipeline_id].append(line[8:])

    def clusterflow_commands_table(self):
        """Make a table of the Cluster Flow commands"""

        # I wrote this when I was tired. Sorry if it's incomprehensible.

        desc = """Every Cluster Flow run will have many different commands.
            MultiQC splits these by whitespace, collects by the tool name
            and shows the first command found. Any terms not found in <em>all</em> subsequent
            calls are replaced with <code>[variable]</code>
            <em>(typically input and ouput filenames)</em>. Each column is for one Cluster Flow run."""

        # Loop through pipelines
        tool_cmds = dict()
        headers = dict()
        for pipeline_id, commands in self.clusterflow_commands.items():
            headers[pipeline_id] = {"scale": False}
            self.var_html = '<span style="background-color:#dedede; color:#999;">[variable]</span>'
            tool_cmd_parts = dict()
            for cmd in commands:
                s = cmd.split()
                tool = self._guess_cmd_name(s)
                if tool not in tool_cmd_parts.keys():
                    tool_cmd_parts[tool] = list()
                tool_cmd_parts[tool].append(s)

            for tool, cmds in tool_cmd_parts.items():
                cons_cmd = self._replace_variable_chunks(cmds)
                # Try again with first two blocks if all variable
                variable_count = cons_cmd.count(self.var_html)
                if variable_count == len(cmds[0]) - 1 and len(cmds[0]) > 2:
                    for subcmd in set([x[1] for x in cmds]):
                        sub_cons_cmd = self._replace_variable_chunks([cmd for cmd in cmds if cmd[1] == subcmd])
                        tool = f"{tool} {subcmd}"
                        if tool not in tool_cmds:
                            tool_cmds[tool] = dict()
                        tool_cmds[tool][pipeline_id] = '<code style="white-space:nowrap;">{}</code>'.format(
                            " ".join(sub_cons_cmd)
                        )
                else:
                    if tool not in tool_cmds:
                        tool_cmds[tool] = dict()
                    tool_cmds[tool][pipeline_id] = '<code style="white-space:nowrap;">{}</code>'.format(
                        " ".join(cons_cmd)
                    )

        table_config = {
            "namespace": "Cluster Flow",
            "id": "clusterflow-commands-table",
            "table_title": "Cluster Flow Commands",
            "col1_header": "Tool",
            "sortRows": False,
            "no_violin": True,
        }
        self.add_section(
            name="Commands",
            anchor="clusterflow-commands",
            description=desc,
            plot=table.plot(tool_cmds, headers, table_config),
        )

    def _replace_variable_chunks(self, cmds):
        """List through a list of command chunks. Return a single list
        with any variable bits blanked out."""

        cons_cmd = None
        while cons_cmd is None:
            for cmd in cmds:
                if cons_cmd is None:
                    cons_cmd = cmd[:]
                else:
                    for idx, s in enumerate(cons_cmd):
                        if s not in cmd:
                            cons_cmd[idx] = self.var_html
        return cons_cmd

    def _guess_cmd_name(self, cmd):
        """Manually guess some known command names, where we can
        do a better job than the automatic parsing."""

        # zcat to bowtie
        if cmd[0] == "zcat" and "bowtie" in cmd:
            return "bowtie"
        # samtools
        if cmd[0] == "samtools":
            return " ".join(cmd[0:2])
        # java (eg. picard)
        if cmd[0] == "java":
            jars = [s for s in cmd if ".jar" in s]
            return os.path.basename(jars[0].replace(".jar", ""))
        return cmd[0]

    def parse_clusterflow_runfiles(self, f):
        """Parse run files generated by Cluster Flow.
        Currently gets pipeline IDs and associated steps."""

        data = dict()
        in_comment = False
        seen_pipeline = False
        cf_file = False
        for line in f["f"]:
            line = line.rstrip()
            # Check that this is from Cluster Flow
            if "Cluster Flow" in line:
                cf_file = True
            # Header
            if line.startswith("Pipeline: "):
                data["pipeline_name"] = line[10:]
            if line.startswith("Pipeline ID: "):
                data["pipeline_id"] = line[13:]
            if line.startswith("Created at "):
                data["pipeline_start"] = line[11:]
            # Config settings
            if line.startswith("@"):
                s = line.split(None, 1)
                key = s[0].replace("@", "").strip()
                try:
                    data[key] = "\t".join(s[1:])
                except IndexError:
                    data[key] = True
            # Comments
            if line.startswith("/*"):
                in_comment = True
            if line.startswith("*/"):
                in_comment = False
            if in_comment:
                if "comment" not in data:
                    data["comment"] = ""
                data["comment"] += line + "\n"
            # Pipeline steps
            if line.strip().startswith("#"):
                if "pipeline_steps" not in data:
                    data["pipeline_steps"] = []
                data["pipeline_steps"].append(line)
                seen_pipeline = True
            # Step output files
            elif seen_pipeline:
                s = line.split("\t")
                if len(s) > 1:
                    if "files" not in data:
                        data["files"] = dict()
                    if s[0] not in data["files"]:
                        data["files"][s[0]] = []
                    data["files"][s[0]].append(s[1:])
        # Parse the start date
        dt = None
        if "pipeline_id" in data:
            s = data["pipeline_id"].split("_")
            dt = datetime.datetime.fromtimestamp(int(s[-1]))
        elif "pipeline_start" in data:
            dt_r = re.match(r"(\d{2}):(\d{2}), (\d{2})-(\d{2})-(\d{4})", data["pipeline_start"])
            if dt_r:
                dt = datetime.datetime(
                    int(dt_r.group(5)),  # year
                    int(dt_r.group(4)),  # month
                    int(dt_r.group(3)),  # day
                    int(dt_r.group(1)),  # hour
                    int(dt_r.group(2)),  # minute
                )

        # Not a Cluster Flow file (eg. Nextflow .run file)
        if not cf_file:
            return None

        if dt is not None:
            data["pipeline_start_dateparts"] = {
                "year": dt.year,
                "month": dt.month,
                "day": dt.day,
                "hour": dt.hour,
                "minute": dt.minute,
                "second": dt.second,
                "microsecond": dt.microsecond,
                "timestamp": time.mktime(dt.timetuple()),
            }
        # Cluster Flow v0.4 and before did not print the pipeline ID in run files
        # Try to guess - will be wrong as no microsecond info, but hopefully unique
        # and reproducible for other run files
        if "pipeline_id" not in data:
            if "pipeline_name" in data and "pipeline_start_dateparts" in data:
                log.debug(f"Trying to guess pipeline ID for file \"{f['fn']}\"")
                data["pipeline_id"] = "cf_{}_{}".format(
                    data["pipeline_name"], data["pipeline_start_dateparts"]["timestamp"]
                )
        return data

    def clusterflow_pipelines_section(self):
        """Generate HTML for section about pipelines, generated from
        information parsed from run files."""
        data = dict()
        pids_guessed = ""
        for f, d in self.clusterflow_runfiles.items():
            pid = d.get("pipeline_id", "unknown")
            if d.get("pipeline_id_guess", False) is True:
                pid += "*"
                pids_guessed = " Project IDs with an asterisk may be inaccurate."
            # Count the number of files going into the first module
            num_starting_files = 0
            for step_name, files in d.get("files", {}).items():
                if step_name.startswith("start"):
                    num_starting_files += len(files)
            # Reformat the date so that column sorting works nicely
            if "pipeline_start_dateparts" in d:
                dt = d["pipeline_start_dateparts"]
                d["pipeline_start"] = "{}-{:02d}-{:02d} {:02d}:{:02d}".format(
                    dt["year"], dt["month"], dt["day"], dt["hour"], dt["minute"]
                )
            if pid not in data:
                data[pid] = d
                data[pid]["num_starting_files"] = int(num_starting_files)
            else:
                data[pid]["num_starting_files"] += int(num_starting_files)

        headers = {
            "pipeline_name": {"title": "Pipeline Name"},
            "pipeline_start": {
                "title": "Date Started",
                "description": "Date and time that pipeline was started (YYYY-MM-DD HH:SS)",
                "scale": False,
            },
            "genome": {
                "title": "Genome ID",
                "description": "ID of reference genome used",
            },
            "num_starting_files": {
                "title": "# Starting Files",
                "format": "{:,.0f}",
                "description": "Number of input files at start of pipeline run.",
                "scale": False,
            },
        }
        table_config = {
            "namespace": "Cluster Flow",
            "id": "clusterflow-pipelines-table",
            "table_title": "Cluster Flow Pipelines",
            "col1_header": "Pipeline ID",
            "no_violin": True,
            "save_file": True,
        }
        self.add_section(
            name="Pipelines",
            anchor="clusterflow-pipelines",
            description=f"Information about pipelines is parsed from <code>*.run</code> files. {pids_guessed}",
            plot=table.plot(data, headers, table_config),
            content=self.clusterflow_pipelines_printout(),
        )

    def clusterflow_pipelines_printout(self):
        """Print the steps used in each Cluster Flow pipeline"""
        data = dict()
        html = ""
        for f, d in self.clusterflow_runfiles.items():
            pid = d.get("pipeline_id", "unknown")
            data[pid] = [d.get("pipeline_name"), "\n".join(d.get("pipeline_steps", []))]
        for pid, d in data.items():
            html += """
                <div class="panel panel-default">
                    <div class="panel-heading"><h3 class="panel-title">Pipeline Steps: {} (<code>{}</code>)</h3></div>
                    <pre class="panel-body" style="border:0; background-color:transparent; padding:0 15px; margin:0; color:#666; font-size:90%;">{}</pre>
                </div>
                """.format(pid, d[0], d[1])
        return html



""" MultiQC submodule to parse output from Picard GcBiasMetrics """

import logging

from multiqc.modules.picard import util
from multiqc.plots import linegraph

# Initialise the logger
log = logging.getLogger(__name__)


def parse_reports(module):
    """
    Find Picard GcBiasMetrics reports and parse their data. There are two types of
    GC bias files:
    * detail: one set of metrics per GC percentage,
    * summary: one set of metrics per input or sample.
    In Picard, they are generated separately with GcBiasDetailMetrics and
    GcBiasSummaryMetrics tools correspondingly. In Sentieon, they are produced with
    the same command, and the summary is only written when the `--summary` option
    is provided.
    """

    data_by_sample = dict()
    summary_data_by_sample = dict()

    # Go through logs and find Metrics
    for f in module.find_log_files("picard/gcbias", filehandles=True):
        # Sample name from input file name by default.
        s_name = f["s_name"]
        gc_col = None
        cov_col = None

        for line in f["f"]:
            maybe_s_name = util.extract_sample_name(
                module,
                line,
                f,
                picard_tool="CollectGcBiasMetrics",
                sentieon_algo="GCBias",
            )
            if maybe_s_name:
                s_name = maybe_s_name

            if s_name is None:
                continue

            if util.is_line_right_before_table(
                line, picard_class=["GcBiasDetailMetrics", "GcBiasSummaryMetrics"], sentieon_algo="GCBias"
            ):
                # Get header - find columns with the data we want
                line = f["f"].readline()
                keys = line.strip("\n").split("\t")

                if "GC" in keys and "NORMALIZED_COVERAGE" in keys:
                    # Detail metrics: one line per GC percentage
                    if s_name in data_by_sample:
                        log.debug(f"Duplicate sample name found in {f['fn']}! Overwriting: {s_name}")
                    data_by_sample[s_name] = dict()
                    gc_col = keys.index("GC")
                    cov_col = keys.index("NORMALIZED_COVERAGE")

                elif "ACCUMULATION_LEVEL" in keys and "GC_DROPOUT" in keys:
                    # Summary metrics - just one line below the header
                    if s_name in summary_data_by_sample:
                        log.debug(f"Duplicate sample name found in {f['fn']}! Overwriting: {s_name}")
                    summary_data_by_sample[s_name] = dict()
                    vals = f["f"].readline().rstrip("\n").split("\t")
                    if len(keys) != len(vals):
                        s_name = None
                        continue

                    for k, v in zip(keys, vals):
                        try:
                            summary_data_by_sample[s_name][k] = float(v)
                        except ValueError:
                            summary_data_by_sample[s_name][k] = v

            elif gc_col is not None and cov_col is not None:
                try:
                    # Note that GC isn't always the first column.
                    s = line.strip("\n").split("\t")
                    data_by_sample[s_name][int(s[gc_col])] = float(s[cov_col])
                except IndexError:
                    gc_col = None
                    cov_col = None

        for s_name in set(data_by_sample.keys()) | set(summary_data_by_sample.keys()):
            module.add_data_source(f, s_name, section="GcBiasMetrics")

    for s_name in list(data_by_sample.keys()):
        if len(data_by_sample[s_name]) == 0:
            data_by_sample.pop(s_name, None)
            log.debug(f"Removing {s_name} as no data parsed")

    for s_name in list(summary_data_by_sample.keys()):
        if len(summary_data_by_sample[s_name]) == 0:
            summary_data_by_sample.pop(s_name, None)
            log.debug(f"Removing {s_name} as no data parsed")

    # Filter to strip out ignored sample names
    data_by_sample = module.ignore_samples(data_by_sample)
    summary_data_by_sample = module.ignore_samples(summary_data_by_sample)

    n_samples = len(data_by_sample.keys() | summary_data_by_sample.keys())
    if n_samples == 0:
        return 0

    # Superfluous function call to confirm that it is used in this module
    # Replace None with actual version if it is available
    module.add_software_version(None)

    if data_by_sample:
        # Plot the graph
        pconfig = {
            "id": f"{module.anchor}_gcbias_plot",
            "title": f"{module.name}: GC Coverage Bias",
            "ylab": "Normalized Coverage",
            "xlab": "% GC",
            "xmin": 0,
            "xmax": 100,
            "xDecimals": False,
            "ymin": 0,
            "yCeiling": 10,
            "tt_label": "<b>{point.x} %GC</b>: {point.y:.2f}",
            "yPlotLines": [
                {"value": 1, "color": "#999999", "width": 2, "dashStyle": "LongDash"},
            ],
        }
        module.add_section(
            name="GC Coverage Bias",
            anchor=f"{module.anchor}-gcbias",
            description="This plot shows bias in coverage across regions of the genome with varying GC content."
            " A perfect library would be a flat line at <code>y = 1</code>.",
            plot=linegraph.plot(data_by_sample, pconfig),
        )

    if summary_data_by_sample:
        # Write parsed summary data to a file
        module.write_data_file(summary_data_by_sample, f"multiqc_{module.anchor}_gcbias")

    # Return the number of detected samples to the parent module
    return n_samples



""" MultiQC module to parse output from Librarian """

import logging

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import heatmap
from multiqc.utils import mqc_colour

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialse the parent object
        super(MultiqcModule, self).__init__(
            name="Librarian",
            anchor="librarian",
            href="https://github.com/DesmondWillowbrook/Librarian",
            info=" - a tool to predict the sequencing library type from the base composition of a FastQ file.",
            doi="10.12688/f1000research.125325.1",
        )

        # To store the summary data
        self.librarian_data = dict()

        # Parse the output files
        self.parse_librarian_data()

        # Remove filtered samples
        self.librarian_data = self.ignore_samples(self.librarian_data)

        # Write the data files to disk
        if not self.librarian_data:
            raise ModuleNoSamplesFound
        log.info(f"Found {len(self.librarian_data)} samples")

        if self.librarian_data:
            self.write_data_file(self.librarian_data, "multiqc_librarian_data")

        self.plot_librarian_heatmap()
        show_general_stats = getattr(config, "librarian", {}).get("show_general_stats", False)
        if show_general_stats:
            self.add_general_stats()

    def parse_librarian_data(self):
        """Loop through Librarian files and parse their data"""
        for f in self.find_log_files("librarian", filehandles=True):
            headers = f["f"].readline().strip().split("\t")
            for line in f["f"]:
                data = dict(zip(headers, line.strip().split("\t")))
                s_name = self.clean_s_name(data["sample_name"], f)
                del data["sample_name"]
                data_float = {}
                for k, v in data.items():
                    try:
                        data_float[k] = float(v)
                    except Exception:
                        data_float[k] = v
                if s_name in self.librarian_data:
                    log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
                self.librarian_data[s_name] = data_float
                self.add_data_source(f, s_name=s_name)

                # Superfluous function call to confirm that it is used in this module
                # Replace None with actual version if it is available
                self.add_software_version(None, s_name)

    def plot_librarian_heatmap(self):
        """Make the heatmap plot"""
        # Get all library types
        hm_library_types = set()
        for d in self.librarian_data.values():
            hm_library_types.update(list(d.keys()))
        hm_library_types = sorted(list(hm_library_types))
        hm_data = []
        hm_sample_names = []
        for s_name in self.librarian_data:
            hm_sample_names.append(s_name)
            sample_data = []
            for library_type in hm_library_types:
                try:
                    sample_data.append(self.librarian_data[s_name][library_type])
                except KeyError:
                    sample_data.append(None)
            hm_data.append(sample_data)

        pconfig = {
            "title": "Librarian: Library Predictions",
            "xTitle": "Library type",
            "yTitle": "Sample name",
            "min": 0,
            "max": 100,
            "square": False,
            "xcats_samples": False,
            "ycats_samples": True,
            "colstops": [[0, "#FFFFFF"], [1, "#FF0000"]],
            "decimalPlaces": 0,
        }

        self.add_section(
            name="Library Type Prediction",
            anchor="librarian-library-type",
            description="""
                For each projected test library, the location on the Compositions/Probability Map is determined.
                This plot shows how published library types are represented at the same location.
            """,
            helptext="""
                Some regions on the map are very specific to a certain library type,
                others are more mixed. Therefore, for some test libraries the results
                will be much clearer than for others.

                The different plots are intended to provide a good overview of how similar
                the test library is to published data.

                The cause of any deviations should be inspected;
                the interpretation will be different depending on how characteristic
                the composition signature of the library type and how far off the
                projection of the test sample is.
            """,
            plot=heatmap.plot(hm_data, hm_library_types, hm_sample_names, pconfig),
        )

    def add_general_stats(self):
        """Add general stats column for most likely library type"""
        # Get most likely library type for each sample
        data = dict()
        lib_types = set()
        for s_name, d in self.librarian_data.items():
            lib_type = max(d, key=d.get)
            score = float(d[lib_type])
            data[s_name] = {
                "most_likely_library_type": lib_type,
                "most_likely_library_type_score": score,
            }
            lib_types.add(lib_type)

        # Generate colours for library types
        lib_types = list(lib_types)
        lib_type_colours = mqc_colour.mqc_colour_scale("Paired", 0, len(lib_types))
        bgcols = {}
        for idx, lib_type in enumerate(lib_types):
            bgcols[lib_type] = lib_type_colours.get_colour(idx)

        headers = {
            "most_likely_library_type": {
                "title": "Likely Type",
                "description": "Most likely library type.",
                "bgcols": bgcols,
            },
            "most_likely_library_type_score": {
                "title": "Type score",
                "description": "Library prediction type score",
                "format": "{:,.0f}",
                "scale": "RdYlGn",
                "min": 0,
                "max": 100,
            },
        }
        self.general_stats_addcols(data, headers)



import logging

from multiqc.modules.base_module import BaseMultiqcModule
from multiqc.modules.dragen.utils import Metric, make_headers
from multiqc.plots import table

log = logging.getLogger(__name__)

METRIC_NAMES = [
    "Unique cell-barcodes",
    "UMI threshold for passing cells",
    "Passing cells",
    "Median reads per cell",
    "Median UMIs per cell",
    "Median genes per cell",
    "Total genes detected",
]
METRICS = [
    Metric(
        id=m,
        title=" ".join([w.title() if w.islower() else w for w in m.split()]),
        in_genstats="#",
        in_own_tabl="#",
        precision=0,
        descr=m,
    )
    for m in METRIC_NAMES
]


class DragenScRnaMetrics(BaseMultiqcModule):
    def add_sc_rna_metrics(self):
        data_by_sample = dict()

        for f in self.find_log_files("dragen/sc_rna_metrics"):
            data = parse_scrna_metrics_file(f)
            s_name = f["s_name"]
            if s_name in data_by_sample:
                log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
            self.add_data_source(f, section="sc_rna_metrics")
            data_by_sample[s_name] = data

            # Superfluous function call to confirm that it is used in this module
            # Replace None with actual version if it is available
            self.add_software_version(None, s_name)

        # Filter to strip out ignored sample names:
        data_by_sample = self.ignore_samples(data_by_sample)

        if not data_by_sample:
            return set()

        # TODO: No test data for this file. No idea if we need this or not.
        # self.write_data_file(data_by_sample, "dragen_sc_rna_metrics")

        gen_stats_headers, table_headers = make_headers(METRIC_NAMES, METRICS)

        self.general_stats_addcols(data_by_sample, gen_stats_headers)
        self.add_section(
            name="Single-Cell RNA Metrics",
            anchor="sc-rna-metrics",
            description="""
            Summary metrics for single-cell RNA.
            """,
            plot=table.plot(
                {
                    sample_name: {
                        metric_name: int(stat) for metric_name, stat in metric.items() if metric_name in METRIC_NAMES
                    }
                    for sample_name, metric in data_by_sample.items()
                },
                table_headers,
                pconfig={
                    "namespace": "Single-Cell RNA Metrics",
                    "id": "dragen-sc-rna-metrics-table",
                },
            ),
        )

        return data_by_sample.keys()


def parse_scrna_metrics_file(f):
    """
    sample.scRNA.metrics.csv

    RUN TIME,,Time loading reference,00:01:31.289,91.29
    RUN TIME,,Time aligning reads,00:00:25.190,25.19
    RUN TIME,,Time duplicate marking,00:00:01.817,1.82
    RUN TIME,,Time sorting and marking duplicates,00:00:07.368,7.37
    RUN TIME,,Time DRAGStr calibration,00:00:07.069,7.07
    """
    data = {}
    for line in f["f"].splitlines():
        tokens = line.split(",")
        if len(tokens) == 4:
            analysis, _, metric, stat = tokens
            percentage = None
        elif len(tokens) == 5:
            analysis, _, metric, stat, percentage = tokens
        else:
            raise ValueError(f"Unexpected number of tokens in line {line}")

        try:
            stat = float(stat)
        except ValueError:
            pass
        data[metric] = stat

    return data



""" MultiQC module to parse output from qc3C """


import itertools
import json
import logging
import os
import re
from collections import defaultdict

import numpy as np

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, linegraph, table

log = logging.getLogger(__name__)

# A 64 color palette defined in four gradient blocks. Red->Yellow, Yellow->Green, Green->Blue, Blue->Deep Blue
# The method color_picker uses these to sample a gradient with varying sparseness

# red to yellow
ry_1 = [
    "#EF476F",
    "#F0506E",
    "#F1586E",
    "#F2616D",
    "#F36A6D",
    "#F4726C",
    "#F57B6C",
    "#F6836B",
    "#F78C6B",
    "#F8956A",
    "#F99D69",
    "#FAA669",
    "#FBAF68",
    "#FCB768",
    "#FDC067",
    "#FEC867",
    "#FFD166",
]

# yellow to green
yg_2 = [
    "#FFD166",
    "#EFD16A",
    "#E0D26D",
    "#D0D271",
    "#C1D275",
    "#B1D378",
    "#A2D37C",
    "#92D37F",
    "#83D483",
    "#73D487",
    "#63D48A",
    "#54D48E",
    "#44D592",
    "#35D595",
    "#25D599",
    "#16D69C",
    "#06D6A0",
]

# green to blue
gb_3 = [
    "#06D6A0",
    "#07D1A1",
    "#07CDA2",
    "#08C8A3",
    "#09C3A5",
    "#09BEA6",
    "#0ABAA7",
    "#0BB5A8",
    "#0CB0A9",
    "#0CABAA",
    "#0DA7AB",
    "#0EA2AC",
    "#0E9DAE",
    "#0F98AF",
    "#1094B0",
    "#108FB1",
    "#118AB2",
]

# blue to dark blue
bd_4 = [
    "#118AB2",
    "#1085AB",
    "#107FA4",
    "#0F7A9E",
    "#0E7597",
    "#0E7090",
    "#0D6A89",
    "#0C6582",
    "#0C607C",
    "#0B5B75",
    "#0A556E",
    "#0A5067",
    "#094B60",
    "#08465A",
    "#084053",
    "#073B4C",
]

# the 64-color gradient
grad64 = ry_1[:-1] + yg_2[:-1] + gb_3[:-1] + bd_4

# 9-color transition related to base colors above.
pal_8 = ["#66C2A5", "#FC8D62", "#8DA0CB", "#E78AC3", "#A6D854", "#FFD92F", "#E5C494", "#B3B3B3"]
rev_8 = pal_8[::-1]


def color_picker(degen):
    """
    Select colours from a 64 colour gradient, where an effort is made to use the full
    width of the gradient depending on the number of elements required.
    :param degen: a list of degeneracies (number of colours per object)
    :return: a flat list of colours equal the sum of degen
    """
    if len(degen) == 1:
        # a single non-ambiguous enzyme, lets make this blue
        if degen[0] not in {1, 4, 16}:
            raise ValueError(f"got {degen[0]} junc_degen values can only be 1, 4 or 16")
        return grad64[0 :: 64 // degen[0]]
    else:
        cols = []
        for n, jd in enumerate(degen):
            if jd not in {1, 4, 16}:
                raise ValueError(f"got {jd} when junc_degen values can only be 1, 4 or 16")
            cols += grad64[16 * n : 16 * n + 16 : 16 // jd]
        return cols


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="qc3C",
            anchor="qc3C",
            href="http://github.com/cerebis/qc3C",
            info="provides reference-free and BAM based quality control for Hi-C data",
            doi="10.1371/journal.pcbi.1008839",
        )

        self.qc3c_data = defaultdict(dict)
        # additional members for conditional plotting of per-genotype junction frequency
        self.do_digest_plot = False
        self.digest_junctions = defaultdict(dict)

        for f in self.find_log_files("qc3C", filehandles=True):
            self.parse_qc3c_log(f)

        for k in self.qc3c_data:
            self.qc3c_data[k] = self.ignore_samples(self.qc3c_data[k])

        # check that any non-empty records were stored under one of the two analysis modes
        n_reports = len(self.qc3c_data["kmer"]) + len(self.qc3c_data["bam"])
        if n_reports == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {n_reports} reports")

        if len(self.qc3c_data["bam"]) > 0:
            self.write_data_file(self.qc3c_data["bam"], "multiqc_qc3c_bam")
            log.debug(f"Found {len(self.qc3c_data['bam'])} BAM analysis reports")

            self.add_section(
                name="BAM mode analysis details",
                anchor="qc3C-bam-runtime-parameters",
                description="""
                    This table details various alignment features which are potentially of interest to
                    researchers attempting to assess the quality of a Hi-C library.
                """,
                plot=self.bam_runtime_table(),
            )

            # self.add_section(
            #     name = 'BAM mode read-pair breakdown',
            #     anchor = 'qc3C-bam-hic-fraction',
            #     description = """
            #         This table details various alignment features which are potentially of interest to
            #         researchers attempting to assess the quality of a Hi-C library.
            #     """,
            #     helptext = """
            #         Here, the **adjusted fraction of read-through** events can be taken as a estimate of the fraction of
            #         Hi-C read-pairs within the library.
            #
            #         * _trans_ (inter-reference) and _cis_ (intra-reference) pairs.
            #         * The fraction of _cis_ pairs whose separation is less than 1000 bp.
            #         * The fraction of total insert extent which was unobserable.
            #         * The fraction of accepted reads whose alignment began with a cutsite.
            #         * The fraction of accepted reads whose alignment ended in a cutsite.
            #         * The fraction of accepted reads which fully aligned and ended in a cutsite.
            #         * The fraction of reads which contained a suspected read-through event (across a junction)
            #         * The fraction of read-through events whose 3-prime end further aligned elsewhere.
            #         * The fraction of read-through events adjusted for unobservable extent.
            #     """,
            #     plot = self.bam_signal_table()
            # )

            self.add_section(
                name="BAM mode read parsing",
                anchor="qc3C-bam-acceptance-plot",
                description="""
                    This figure displays a breakdown of proportion of parsed reads rejected due to various
                    criteria and the proportion that were accepted.
                """,
                plot=self.bam_acceptance_plot(),
            )

            # the following table is disabled to simplify report, as it is redundant consider the plot that follows.
            #
            # self.add_section(
            #     name = 'BAM mode HiCPro categories',
            #     anchor = 'qc3C-bam-hicpro-table',
            #     description = """
            #         This table details the read-pair categorisation devised by
            #         [HiC-Pro](https://github.com/nservant/HiC-Pro).
            #     """,
            #     helptext = """
            #         As the field has moved from 6-cutter to 4-cutter enzymes, and subsequently dual-enzyme digests,
            #         the higher density of sites has made this framework less useful, since it has become increasingly
            #         easy to satisfy the intervening site criteria.
            #     """,
            #     plot = self.bam_hicpro_table()
            # )

            self.add_section(
                name="BAM mode HiC-Pro validation",
                anchor="qc3C-bam-valid-plot",
                description="""
                    A visualisation of the read-pair categories devised by
                    [HiC-Pro](https://github.com/nservant/HiC-Pro).
                """,
                helptext="""
                    As the field has moved from 6-cutter to 4-cutter enzymes, and subsequently dual-enzyme digests, the
                    higher density of sites has made this framework less useful, since it has become increasingly easy
                    to satisfy the intervening site criteria.
                """,
                plot=self.bam_valid_plot(),
            )

            self.add_section(
                name="BAM mode long-range pairs",
                anchor="qc3C-bam-longrange-plot",
                description="This plot visualises the breakdown of read-pairs based on separation distance.",
                helptext="""
                    The breakdown of separation distance is only calculated for *cis*-mapping pairs.

                    Ideally, Hi-C proximity ligation should produce many pairs which are greater than 1000 bp apart.
                    However, these statistics are strongly influenced by the state of the reference. For draft assemblies
                    the distance at which pairs can map is limited by the degree of fragmentation and length of contigs.
                    As a result, many more pairs will be categorised as *trans*-mapping and pairs which are truly
                    inter-molecular cannot be distinguished from those which are merely inter-contig.
                """,
                plot=self.bam_longrange_plot(),
            )

            self.add_section(
                name="BAM mode distribution of fragment separation",
                anchor="qc3C-bam-fragment-histogram",
                description="""
                    This figure displays the a normalised histogram of read-pair separation binned uniformly in
                    log-space.

                    Due to the binning strategy, the x-axis is log-scaled and visually accommodates pair separations up
                    to 1 million bp. The inferred insert size for each library is represented by a dashed, grey vertical
                    line. The y-axis is log-scaled by default, allowing the density attributed to long-range pairs to be
                    more easily seen.
                """,
                helptext="""
                    A characteristic of Hi-C libraries, is the presence of a large peak below 1000 bp. qc3C attributes
                    this to regular (and undesirable) shotgun pairs creeping through the Hi-C protocol. The peak is used
                    by qc3C to infer the insert size, which is later employed to estimate unobservable extent of
                    inserts.

                    **Note:** the inferred insert size can be significantly smaller than what a sequencing facility
                    might report the experimentally determined insert size to be. This discrepancy can be explained by
                    the failure to account for the additional adapter sequence when fragments are assessed during
                    library preparation.
                """,
                plot=self.bam_fragment_histogram(),
            )

            if self.do_digest_plot:
                self.add_section(
                    name="BAM mode junction breakdown",
                    anchor="qc3C-bam-junction-plot",
                    description="""
                        This figure displays the frequency at which a library's possible junction sequences
                        are actually observed in the reads. (_Trivial single-digests are ignored_)
                    """,
                    helptext="""
                        For trivial single-enzyme digests, there is only one possible junction sequence and so the
                        result for these experiments are not plotted. For dual-enzyme (such as Phase Genomics) there are
                        four potential junctions, while for dual-enzyme digests with one ambiguous site (such as Arima
                        Genomics) there are 16 possible junction sequences.

                        How efficiently the more complicated library protocols are at producing hybrid junctions is
                        possibly just a point of interest.

                        Junctions are named for which enzymes was responsible for creating the 5' and 3' ends.
                        E.g. `Sau3AI/MluCI` would involve two different enzymes, while `Sau3AI/Sau3AI` only one, as
                        would be the case in a single-enzyme digest. Proceeding the name is the actual junction
                        sequence.

                        The junctions are grouped by their 5' and then 3' enzyme, while the color spectrum used across
                        each bar aims to emphasise these enzymatic sources.

                        **Note:** in BAM mode, the counts **are** controlled for false positives, in the sense that read
                        alignments must terminate at a cutsite, but the read sequence must continue and contain the
                        observed junction.
                    """,
                    plot=self.bam_junction_plot(),
                )

        if len(self.qc3c_data["kmer"]) > 0:
            self.write_data_file(self.qc3c_data["kmer"], "multiqc_qc3c_kmer")
            log.debug(f"Found {len(self.qc3c_data['kmer'])} k-mer analysis reports")

            self.add_section(
                name="K-mer mode runtime details",
                anchor="qc3C-kmer-runtime-parameters",
                description="""
                    This table includes user specified input options, observed read-length and unobservable fraction.
                """,
                plot=self.kmer_runtime_table(),
            )

            self.add_section(
                name="K-mer mode Hi-C fraction",
                anchor="qc3C-kmer-hic-fraction",
                description="This table lists the inferred proportion of Hi-C proximity ligation fragments.",
                helptext="""
                    Here, **Mean adjusted Hi-C fraction** represents the best estimate of the proportion of a library's
                    read-pairs which are a product of proximity ligation. This figure is arrived at by correcting the
                    raw estimate for the fraction of insert extent which was not observable.

                    The observable extent is limited by the length of reads relative to the supplied insert size, as
                    well as a further constraint on flanking sequence around any suspected junction sequence.
                """,
                plot=self.kmer_signal_table(),
            )

            self.add_section(
                name="K-mer mode read parsing",
                anchor="qc3C-kmer-acceptance-plot",
                description="""
                    This figure displays a breakdown of proportion of parsed reads rejected due to various
                    criteria and the proportion that were accepted.
                """,
                plot=self.kmer_acceptance_plot(),
            )

            # TODO this figure seems of low value
            # self.add_section(
            #     name = 'K-mer mode junction proportion',
            #     anchor = 'qc3C-kmer-signal-plot',
            #     description = """
            #         This figure displays the proportion of inspected reads which contained a putative junction
            #         sequence, as compared against all reads inspected.
            #     """,
            #     plot = self.kmer_signal_plot()
            # )

            # TODO consider whether this figure should be displayed when there are non-trivial digests
            if self.do_digest_plot:
                self.add_section(
                    name="K-mer mode junction breakdown",
                    anchor="qc3C-kmer-junction-plot",
                    description="""
                        This figure displays the frequency at which a library's possible junction sequences
                        are actually observed in the reads. (_Trivial single-digests are ignored_)
                    """,
                    helptext="""
                        For trivial single-enzyme digests, there is only one possible junction sequence and so the
                        result for these experiments are not plotted. For dual-enzyme (such as Phase Genomics) there are
                        four potential junctions, while for dual-enzyme digests with one ambiguous site (such as Arima
                        Genomics) there are 16 possible junction sequences.

                        How efficiently the more complicated library protocols are at producing hybrid junctions is
                        possibly just a point of interest.

                        Junctions are named for which enzymes was responsible for creating the 5' and 3' ends.
                        E.g. `Sau3AI/MluCI` would involve two different enzymes, while `Sau3AI/Sau3AI` only one, as
                        would be the case in a single-enzyme digest. Proceeding the name is the actual junction
                        sequence.

                        The junctions are grouped by their 5' and then 3' enzyme, while the color spectrum used across
                        each bar aims to emphasise these enzymatic sources.

                    **Note:** in k-mer mode, the counts are not controlled for false positives.
                    """,
                    plot=self.kmer_junction_plot(),
                )

    @staticmethod
    def _drop_time(s):
        """remove time"""
        m = re.match(r"(.*) .+$", s)
        return s if m is None else m.group(1)

    @staticmethod
    def _drop_name(s):
        """drop leading program name from version string"""
        return s.split()[-1]

    def bam_runtime_table(self):
        config = {
            "id": "qc3C_bam_runtime_table",
            "namespace": "qc3C",
            "col1_header": "Sample",
            "scale": False,
            "title": "qc3C: BAM mode runtime parameters",
        }

        headers = {
            "b_run_timestamp": {
                "title": "Date",
                "description": "Analysis time stamp",
                "modify": MultiqcModule._drop_time,
                "hidden": True,
            },
            "b_mode": {"title": "Run Mode", "description": "Analysis mode used", "hidden": True},
            "b_min_mapq": {
                "title": "Min MapQ",
                "description": "Minimum accepted mapping quality",
                "min": 0,
                "format": "{:d}",
                "scale": False,
            },
            "b_enzymes": {"title": "Digest", "description": "Enzymes used in digest"},
            # 'b_seed': {
            #     'title': 'Seed',
            #     'description': 'Random seed',
            #     'format': '{:d}',
            #     'scale': False,
            #     'hidden': True,
            # },
            # 'b_max_obs': {
            #     'title': 'Max obs',
            #     'description': 'User specified maximum number of observations',
            #     'min': 0,
            #     'format': '{:,d}',
            #     'scale': 'OrRd',
            #     'modify': lambda x: 'n/a' if x == -1 else x
            # },
            "b_n_accepted_pairs": {
                "title": "Accepted pairs",
                "description": "Number of pairs accepted for analysis",
                "min": 0,
                "format": "{:,d}",
                "scale": "Oranges",
            },
            # 'b_sample_rate': {
            #     'title': 'Sample rate',
            #     'description': 'Sub-sampling probability',
            #     'min': 0,
            #     'max': 1,
            #     'format': '{:g}',
            #     'scale': 'Greys'
            # },
            # 'b_obs_insert_median': {
            #     'title': 'Insert median',
            #     'description': 'Estimated median insert size',
            #     'min': 0,
            #     'format': '{:,.0f}',
            #     'suffix': 'bp',
            #     'scale': 'Greens'
            # },
            "b_mean_readlen": {
                "title": "Read length",
                "description": "Average observed read length",
                "format": "{:,.0f}",
                "suffix": "bp",
                "scale": "Blues",
            },
            "b_obs_insert_mean": {
                "title": "Insert length",
                "description": "Inferred insert size",
                "min": 0,
                "format": "{:,.0f}",
                "suffix": "bp",
                "scale": "Purples",
            },
            "b_unobs_fraction": {
                "title": "Unobserved",
                "description": "Fraction of total insert extent that was unobservable",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "Reds",
            },
            "b_p_read_thru": {
                "title": "Read-thru",
                "description": "Fraction of reads whose alignments end in a cutsite and whose sequence continues for the full junction",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "Greens",
            },
        }

        return table.plot(self.qc3c_data["bam"], headers, config)

    def bam_longrange_plot(self):
        config = {
            "id": "qc3C_bam_longrange_plot",
            "namespace": "qc3C",
            "title": "qc3C: BAM mode long-range pairs",
            "ylab": "Number of Reads",
            "cpswitch_counts_label": "Number of Reads",
        }

        categories = {
            "b_n_10k_": {"name": "10000 and further bp", "color": "#118AB2"},
            "b_n_5k_10k": {"name": "from 5000 to 10000 bp", "color": "#0CABAA"},
            "b_n_1k_5k": {"name": "from 1000 to 5000 bp", "color": "#07D1A1"},
            "b_n_short_inserts": {"name": "less than 1000 bp", "color": "#FFD166"},
            "b_n_trans_pairs": {"name": "trans pairs", "color": "#F78C6B"},
        }

        return bargraph.plot(self.qc3c_data["bam"], categories, config)

    def bam_acceptance_plot(self):
        config = {
            "id": "qc3C_bam_acceptance_plot",
            "namespace": "qc3C",
            "title": "qc3C: BAM mode read parsing results",
            "ylab": "Number of Reads",
            "hide_zero_cats": False,
            "cpswitch_counts_label": "Number of Reads",
        }

        categories = {
            "b_n_accepted_reads": {"name": "Accepted", "color": rev_8[-1]},
            "b_n_ref_term_reads": {"name": "Truncated", "color": rev_8[6]},
            "b_n_weak_mapping_reads": {"name": "Weak mapping", "color": rev_8[5]},
            "b_n_supplementary_reads": {"name": "Supplementary", "color": rev_8[4]},
            "b_n_secondary_reads": {"name": "Secondary", "color": rev_8[3]},
            "b_n_low_mapq_reads": {"name": "Low mapq", "color": rev_8[2]},
            "b_n_unmapped_reads": {"name": "Unmapped", "color": rev_8[1]},
            "b_n_skipped_reads": {"name": "Skipped", "color": rev_8[0]},
        }
        return bargraph.plot(self.qc3c_data["bam"], categories, config)

    def bam_signal_table(self):
        config = {"id": "qc3C_bam_signal_table", "namespace": "qc3C", "hide_zero_cats": False, "col1_header": "Sample"}

        headers = {
            "b_unobs_fraction": {
                "title": "Unobservable extent",
                "description": "Estimated fraction of total fragment extent that was unobservable",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "Reds",
            },
            "b_p_read_thru": {
                "title": "Read-thru",
                "description": "Fraction of reads whose alignments end in a cutsite and whose sequence continues for the full junction",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "Greens",
            },
        }
        return table.plot(self.qc3c_data["bam"], headers, config)

    def bam_hicpro_table(self):
        config = {"id": "qc3C_bam_hicpro_table", "namespace": "qc3C", "hide_zero_cats": False, "col1_header": "Sample"}

        headers = {
            "b_p_informative_fr": {"title": "Valid FR", "min": 0, "max": 100, "suffix": "%", "scale": "Greens"},
            "b_p_informative_rf": {"title": "Valid RF", "min": 0, "max": 100, "suffix": "%", "scale": "Greens"},
            "b_p_informative_ffrr": {
                "title": "Valid FF|RR",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "Greens",
            },
            "b_p_uninformative_religation": {
                "title": "Religation",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "Blues",
            },
            "b_p_uninformative_dangling_ends": {
                "title": "Dangling End",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "Reds",
            },
            "b_p_uninformative_self_circle": {
                "title": "Self-circle",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "Reds",
            },
            "b_p_uninformative_ffrr": {
                "title": "Invalid FF|RR",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "Reds",
            },
        }
        return table.plot(self.qc3c_data["bam"], headers, config)

    def bam_valid_plot(self):
        config = {
            "id": "qc3C_bam_valid_plot",
            "namespace": "qc3C",
            "title": "qc3C: BAM mode valid vs invalid HiC-Pro categories",
            "ylab": "Number of Reads",
            "hide_zero_cats": False,
            "cpswitch_counts_label": "Number of Reads",
        }

        categories = {
            "b_n_informative_fr": {"name": "Valid FR", "color": "#41ab5d"},
            "b_n_informative_rf": {"name": "Valid RF", "color": "#74c476"},
            "b_n_informative_ffrr": {"name": "Valid FF|RR", "color": "#a1d99b"},
            "b_n_uninformative_religation": {"name": "Religation", "color": "#fcbba1"},
            "b_n_uninformative_dangling_ends": {"name": "Dangling End", "color": "#fc9272"},
            "b_n_uninformative_self_circle": {"name": "Self-circle", "color": "#fb6a4a"},
            "b_n_uninformative_ffrr": {"name": "Invalid FF|RR", "color": "#ef3b2c"},
        }
        return bargraph.plot(self.qc3c_data["bam"], categories, config)

    def bam_junction_plot(self):
        config = {
            "id": "qc3C_bam_junction_plot",
            "namespace": "qc3C",
            "title": "qc3C: BAM mode read-thru ligation product frequency",
            "ylab": "Number of reads",
            "hide_zero_cats": False,
            "use_legend": False,
        }

        categories = dict()
        for v in self.digest_junctions["bam"].values():
            for vi in v:
                categories[vi["name"]] = vi

        return bargraph.plot(self.qc3c_data["bam"], categories, config)

    def bam_fragment_histogram(self):
        median_lines = []
        for smpl in self.qc3c_data["bam"]:
            median_lines.append(
                {
                    "value": self.qc3c_data["bam"][smpl]["b_obs_insert_median"],
                    "color": "#D8E2DC",
                    "width": 2,
                    "dashStyle": "ShortDashDot",
                }
            )

        config = {
            "id": "qc3C_bam_fragment_histogram",
            "title": "qc3C: BAM mode distribution of pair separation",
            "namespace": "qc3C",
            "cpswitch_counts_label": "Density",
            "logswitch": True,
            "logswitch_active": True,
            "logswitch_label": "Log10 [Density]",
            "xLog": True,
            # 'yLog': True,
            "xPlotLines": median_lines,
            "xlab": "Log10 [Separation]",
            "ylab": "Density",
            "tt_label": "x:{point.x:.0f} bp, y:{point.y:.8f}",
        }

        data = {}
        for smpl in self.qc3c_data["bam"]:
            data[smpl] = self.qc3c_data["bam"][smpl]["frag_hist"]

        return linegraph.plot(data, config)

    def kmer_runtime_table(self):
        config = {
            "id": "qc3C_kmer_runtime_table",
            "namespace": "qc3C",
            "col1_header": "Sample",
            "scale": False,
            "title": "qc3C: K-mer mode runtime parameters",
        }

        headers = {
            "k_run_timestamp": {
                "title": "Date",
                "description": "Analysis time stamp",
                "modify": MultiqcModule._drop_time,
                "hidden": True,
            },
            "k_mode": {"title": "Run Mode", "description": "Analysis mode used", "hidden": True},
            "k_kmer_size": {
                "title": "k",
                "description": "Library k-mer size",
                "min": 0,
                "format": "{:d}",
                "scale": False,
                "hidden": True,
            },
            "k_enzymes": {"title": "Digest", "description": "Enzymes used in digest"},
            "k_n_accepted_reads": {
                "title": "Accepted reads",
                "description": "Number of reads accepted for analysis",
                "min": 0,
                "format": "{:,d}",
                "scale": "BuGn",
            },
            "k_mean_insert": {
                "title": "Insert length",
                "description": "User-specified insert size",
                "min": 0,
                "format": "{:,.0f}",
                "suffix": "bp",
                "scale": "Greens",
            },
            "k_mean_readlen": {
                "title": "Read length",
                "description": "Observed average read length",
                "format": "{:,.0f}",
                "suffix": "bp",
                "scale": "Blues",
            },
            "k_unobs_fraction": {
                "title": "Unobservable extent",
                "description": "Estimated mean of the unobservable portion of fragments",
                "shared_key": "unobs_mean",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "Reds",
            },
        }
        return table.plot(self.qc3c_data["kmer"], headers, config)

    def kmer_signal_table(self):
        config = {
            "id": "qc3C_kmer_signal_table",
            "namespace": "qc3C",
            "col1_header": "Sample",
            "title": "qc3C: K-mer mode Hi-C fraction",
        }

        headers = {
            "k_raw_fraction": {
                "title": "Mean raw Hi-C fraction",
                "description": "Estimated mean of Hi-C fraction from only the observable extent",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "Greens",
            },
            "k_adj_fraction": {
                "title": "Mean adjusted Hi-C fraction",
                "description": "Estimated mean of Hi-C fraction adjusted for unobservable extent",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "Blues",
            },
        }
        return table.plot(self.qc3c_data["kmer"], headers, config)

    def kmer_acceptance_plot(self):
        config = {
            "id": "qc3C_kmer_acceptance_plot",
            "namespace": "qc3C",
            "title": "qc3C: K-mer mode read parsing results",
            "ylab": "Number of Reads",
            "hide_zero_cats": False,
            "cpswitch_counts_label": "Number of Reads",
        }

        categories = {
            "k_n_accepted_reads": {"name": "Accepted", "color": rev_8[-1]},
            "k_n_low_cov": {"name": "Low cov", "color": rev_8[6]},
            "k_n_zero_cov": {"name": "Zero cov", "color": rev_8[5]},
            "k_n_high_cov": {"name": "High cov", "color": rev_8[4]},
            "k_n_ambiguous": {"name": "Ambiguous", "color": rev_8[3]},
            "k_n_no_flank": {"name": "No flank", "color": rev_8[2]},
            "k_n_too_short": {"name": "Too short", "color": rev_8[1]},
            "k_n_skipped": {"name": "Skipped", "color": rev_8[0]},
        }
        return bargraph.plot(self.qc3c_data["kmer"], categories, config)

    def kmer_signal_plot(self):
        config = {
            "id": "qc3C_kmer_signal_plot",
            "namespace": "qc3C",
            "title": "qc3C: K-mer mode signal content",
            "ylab": "Number of Reads",
            "hide_zero_cats": False,
            "stacking": None,
            "cpswitch": False,
            "cpswitch_c_active": False,
            "cpswitch_counts_label": "Number of Reads",
        }

        categories = {
            "k_raw_fraction": {"name": "Raw Hi-C fraction", "color": "#ef3b2c"},
            "k_adj_fraction": {"name": "Adjusted Hi-C fraction", "color": "#41ab5d"},
        }
        return bargraph.plot(self.qc3c_data["kmer"], categories, config)

    def kmer_junction_plot(self):
        config = {
            "id": "qc3C_kmer_frequency_plot",
            "namespace": "qc3C",
            "title": "qc3C: K-mer mode putative ligation product frequency",
            "ylab": "Number of Reads",
            "hide_zero_cats": False,
            "use_legend": False,
        }

        categories = dict()
        for _cat in self.digest_junctions["kmer"].values():
            for vi in _cat:
                categories[vi["name"]] = vi

        return bargraph.plot(self.qc3c_data["kmer"], categories, config)

    def parse_qc3c_log(self, f):
        def _none_to(x, y):
            return y if x is None else y

        try:
            parsed = [json.loads(_l) for _l in f["f"]]
            if len(parsed) > 1:
                log.warning(
                    "Multiple records encountered in qc3C JSON file {}. Only the last report will be used".format(
                        os.path.join(f["root"], f["fn"])
                    )
                )
            parsed = parsed[-1]
        except json.JSONDecodeError:
            log.warning(f"Could not parse qc3C JSON: '{f['fn']}'")
            return

        s_name = self.clean_s_name(os.path.basename(f["root"]), f, root=os.path.dirname(f["root"]))
        if s_name in self.qc3c_data:
            log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")

        # Add version info
        version = parsed["runtime_info"]["qc3C_version"].split()[-1]
        self.add_software_version(version, f["s_name"])

        try:
            analysis_mode = parsed["mode"]
            if "unobs_fraction" not in parsed and "unobs_frac" in parsed:
                parsed["unobs_fraction"] = parsed["unobs_frac"]

            if analysis_mode == "bam":
                if "vs_accepted" not in parsed["separation_bins"] and "zvs_accepted " in parsed["separation_bins"]:
                    parsed["separation_bins"]["vs_accepted"] = parsed["separation_bins"]["zvs_accepted"]

                # set some variables to shorten the lines below
                inf = parsed["classification"]["informative"]
                uninf = parsed["classification"]["uninformative"]
                n_cis_pairs = parsed["n_cis_pairs"]
                n_accepted_pairs = parsed["n_accepted_pairs"]
                n_paired_reads = n_accepted_pairs * 2

                self.qc3c_data["bam"][s_name] = {
                    "b_qc3C_version": parsed["runtime_info"]["qc3C_version"],
                    "b_run_timestamp": parsed["runtime_info"]["run_timestamp"],
                    "b_mode": parsed["mode"],
                    "b_enzymes": ", ".join(parsed["input_args"]["enzymes"]),
                    "b_seed": parsed["input_args"]["seed"],
                    "b_sample_rate": _none_to(parsed["input_args"]["sample_rate"], 1),
                    "b_max_obs": _none_to(parsed["input_args"]["max_obs"], -1),
                    "b_n_skipped_reads": parsed["n_skipped_reads"],
                    "b_n_unmapped_reads": parsed["n_unmapped"],
                    "b_n_analysed_reads": parsed["n_analysed_reads"],
                    "b_n_low_mapq_reads": parsed["n_low_mapq"],
                    "b_n_ref_len_reads": parsed["n_ref_len"],
                    "b_n_secondary_reads": parsed["n_secondary"],
                    "b_n_supplementary_reads": parsed["n_supplementary"],
                    "b_n_weak_mapping_reads": parsed["n_weak_mapping"],
                    "b_n_ref_term_reads": parsed["n_ref_term"],
                    "b_n_accepted_reads": parsed["n_accepted_reads"],
                    "b_obs_insert_mean": parsed["obs_insert_mean"],
                    "b_obs_insert_median": parsed["obs_insert_median"],
                    "b_mean_readlen": parsed["mean_readlen"],
                    "b_n_analysed_pairs": parsed["n_analysed_pairs"],
                    "b_n_accepted_pairs": parsed["n_accepted_pairs"],
                    "b_n_trans_pairs": parsed["n_trans_pairs"],
                    "b_p_trans_pairs": parsed["n_trans_pairs"] / n_accepted_pairs * 100,
                    "b_p_cis_pairs": n_cis_pairs / n_accepted_pairs * 100,
                    "b_unobs_fraction": parsed["unobs_fraction"] * 100,
                    "b_p_cs_start": parsed["digest_stats"]["cs_start"] / n_paired_reads * 100,
                    "b_p_cs_term": parsed["digest_stats"]["cs_term"] / n_paired_reads * 100,
                    "b_p_cs_full": parsed["digest_stats"]["cs_full"] / n_paired_reads * 100,
                    "b_p_read_thru": parsed["digest_stats"]["read_thru"] / n_paired_reads * 100,
                    "b_p_is_split": parsed["digest_stats"]["is_split"] / n_paired_reads * 100,
                    "b_adj_read_thru": parsed["digest_stats"]["read_thru"]
                    / n_paired_reads
                    * 100
                    * 1
                    / (1 - parsed["unobs_fraction"]),
                    "b_n_short_inserts": parsed["n_short_inserts"],
                    "b_p_short_inserts": parsed["n_short_inserts"] / n_accepted_pairs * 100,
                    "b_n_informative_fr": inf["fr"],
                    "b_n_informative_rf": inf["rf"],
                    "b_n_informative_ffrr": inf["ffrr"],
                    "b_n_uninformative_religation": uninf["religation"],
                    "b_n_uninformative_dangling_ends": uninf["dangling_ends"],
                    "b_n_uninformative_self_circle": uninf["self_circle"],
                    "b_n_uninformative_ffrr": uninf["ffrr"],
                    "b_p_informative_fr": inf["fr"] / n_cis_pairs * 100,
                    "b_p_informative_rf": inf["rf"] / n_cis_pairs * 100,
                    "b_p_informative_ffrr": inf["ffrr"] / n_cis_pairs * 100,
                    "b_p_uninformative_religation": uninf["religation"] / n_cis_pairs * 100,
                    "b_p_uninformative_dangling_ends": uninf["dangling_ends"] / n_cis_pairs * 100,
                    "b_p_uninformative_self_circle": uninf["self_circle"] / n_cis_pairs * 100,
                    "b_p_uninformative_ffrr": uninf["ffrr"] / n_cis_pairs * 100,
                    "b_n_1kb": parsed["separation_bins"]["counts"][0],
                    "b_n_5kb": parsed["separation_bins"]["counts"][1],
                    "b_n_10kb": parsed["separation_bins"]["counts"][2],
                    "b_p_1kb_vs_accepted": parsed["separation_bins"]["vs_accepted"][0],
                    "b_p_5kb_vs_accepted": parsed["separation_bins"]["vs_accepted"][1],
                    "b_p_10kb_vs_accepted": parsed["separation_bins"]["vs_accepted"][2],
                    "b_p_1kb_vs_cis": parsed["separation_bins"]["vs_all_cis"][0],
                    "b_p_5kb_vs_cis": parsed["separation_bins"]["vs_all_cis"][1],
                    "b_p_10kb_vs_cis": parsed["separation_bins"]["vs_all_cis"][2],
                    "b_n_1k_5k": parsed["separation_bins"]["counts"][0] - parsed["separation_bins"]["counts"][1],
                    "b_n_5k_10k": parsed["separation_bins"]["counts"][1] - parsed["separation_bins"]["counts"][2],
                    "b_n_10k_": parsed["separation_bins"]["counts"][2],
                }

                fhist = {}
                for x, y in itertools.zip_longest(
                    parsed["separation_histogram"]["mid_points"], parsed["separation_histogram"]["counts"]
                ):
                    fhist[float(x)] = float(y)
                self.qc3c_data["bam"][s_name]["frag_hist"] = fhist

            elif analysis_mode == "kmer":
                for k in "raw_fraction", "adj_fraction", "unobs_fraction":
                    if parsed[k] is None:
                        parsed[k] = "Error - adjusted value would exceed 100"
                    else:
                        parsed[k] = np.array(parsed[k]).mean() * 100

                self.qc3c_data["kmer"][s_name] = {
                    "k_qc3C_version": parsed["runtime_info"]["qc3C_version"],
                    "k_run_timestamp": parsed["runtime_info"]["run_timestamp"],
                    "k_mode": parsed["mode"],
                    "k_kmer_size": parsed["input_args"]["kmer_size"],
                    "k_enzymes": ", ".join(parsed["input_args"]["enzymes"]),
                    "k_seed": parsed["input_args"]["seed"],
                    "k_sample_rate": _none_to(parsed["input_args"]["sample_rate"], 1),
                    "k_max_freq": parsed["input_args"]["max_coverage"],
                    "k_mean_insert": parsed["input_args"]["mean_insert"],
                    "k_max_freq_quantile": parsed["input_args"]["max_freq_quantile"],
                    "k_max_obs": _none_to(parsed["input_args"]["max_obs"], -1),
                    "k_n_skipped": parsed["n_parsed_reads"] - parsed["n_analysed_reads"],
                    "k_n_analysed_reads": parsed["n_analysed_reads"],
                    "k_n_too_short": parsed["n_too_short"],
                    "k_n_no_flank": parsed["n_no_flank"],
                    "k_n_ambiguous": parsed["n_ambiguous"],
                    "k_n_high_cov": parsed["n_high_cov"],
                    "k_n_low_cov": parsed["n_low_cov"],
                    "k_n_zero_cov": parsed["n_zero_cov"],
                    "k_n_accepted_reads": parsed["n_accepted_reads"],
                    "k_n_without_junc": parsed["n_without_junc"],
                    "k_n_with_junc": parsed["n_with_junc"],
                    "k_mean_readlen": parsed["mean_readlen"],
                    "k_n_cs_start": parsed["cs_start"],
                    "k_raw_fraction": parsed["raw_fraction"],
                    "k_adj_fraction": parsed["adj_fraction"],
                    "k_unobs_fraction": parsed["unobs_fraction"],
                }

            # if any experiment contains a digest tht is non-trivial and produces
            # more than one possible junction sequence, prepare the supporting data
            # to render the junction frequency plot
            if len(parsed["junction_frequency"]) > 1:
                self.do_digest_plot = True
                log.debug(f"Enabled junction frequency plot for non-trivial digest: {f['root']}")
                # include the junction frequencies (1 or many depending on digest)
                self.qc3c_data[analysis_mode][s_name].update(parsed["junction_frequency"])

                # calculate the degeneracy of junction sequences per enzymatic combination (5p end =/= 3p end)
                # this can vary due to ambiguous bases in restriction site
                degen_count = {
                    f"{v['enz5p']}/{v['enz3p']}": 4 ** v["junction"].count("N")
                    for k, v in parsed["digestion"]["junctions"].items()
                }
                # get a palette for this series
                cols = color_picker(list(degen_count.values()))
                # sort these in accordance with that above
                juncs = np.sort(
                    np.array(
                        [(k.split(" ")[0], k) for k in parsed["junction_frequency"]],
                        dtype=np.dtype([("a", "S100"), ("b", "S100")]),
                    )
                )

                # keep a record of how these should be colored per sample
                self.digest_junctions[analysis_mode][s_name] = [
                    {"name": juncs[i][1].decode(), "color": cols[i]} for i in range(len(juncs))
                ]

            self.add_data_source(f, s_name, section=analysis_mode)

        except KeyError as ex:
            log.debug(
                "The entry {} was not found in the qc3C JSON file '{}', skipping sample {}".format(
                    str(ex), os.path.join(f["root"], f["fn"]), f["s_name"]
                )
            )
            return



""" MultiQC submodule to parse output from Picard QualityYieldMetrics """

import logging
from collections import OrderedDict

from multiqc import config
from multiqc.modules.picard import util

# Initialise the logger
log = logging.getLogger(__name__)

DESC = OrderedDict(
    [
        ("TOTAL_READS", "The total number of reads in the input file"),
        ("PF_READS", "The number of reads that are PF - pass filter"),
        ("READ_LENGTH", "The average read length of all the reads (will be fixed for a lane)"),
        ("TOTAL_BASES", "The total number of bases in all reads"),
        ("PF_BASES", "The total number of bases in all PF reads"),
        ("Q20_BASES", "The number of bases in all reads that achieve quality score 20 or higher"),
        ("PF_Q20_BASES", "The number of bases in PF reads that achieve quality score 20 or higher"),
        ("Q30_BASES", "The number of bases in all reads that achieve quality score 30 or higher"),
        ("PF_Q30_BASES", "The number of bases in PF reads that achieve quality score 30 or higher"),
        ("Q20_EQUIVALENT_YIELD", "The sum of quality scores of all bases divided by 20"),
        ("PF_Q20_EQUIVALENT_YIELD", "The sum of quality scores of all bases in PF reads divided by 20"),
    ]
)


def parse_reports(module):
    """Find Picard QualityYieldMetrics reports and parse their data"""

    # Set up vars
    data_by_sample = dict()
    expected_header = list(DESC.keys())

    # Go through logs and find Metrics
    for f in module.find_log_files("picard/quality_yield_metrics", filehandles=True):
        # Sample name from input file name by default.
        s_name = f["s_name"]

        for line in f["f"]:
            maybe_s_name = util.extract_sample_name(
                module,
                line,
                f,
                picard_tool="CollectQualityYieldMetrics",
            )
            if maybe_s_name:
                s_name = maybe_s_name

            if s_name is None:
                continue

            if util.is_line_right_before_table(line, picard_class="QualityYieldMetrics"):
                keys = f["f"].readline().strip("\n").split("\t")
                if keys != expected_header:
                    continue

                if s_name in data_by_sample:
                    log.debug(f"Duplicate sample name found in {f['fn']}! Overwriting: {s_name}")
                module.add_data_source(f, s_name, section="QualityYieldMetrics")

                vals = []
                for v in f["f"].readline().strip("\n").split("\t"):
                    try:
                        v = int(v)
                    except ValueError:
                        pass
                    vals.append(v)
                data_by_sample[s_name] = dict(zip(keys, vals))
                s_name = None

    data_by_sample = module.ignore_samples(data_by_sample)
    if not data_by_sample:
        return 0

    # Superfluous function call to confirm that it is used in this module
    # Replace None with actual version if it is available
    module.add_software_version(None)

    # Write parsed data to a file
    module.write_data_file(data_by_sample, f"multiqc_{module.anchor}_QualityYieldMetrics")

    # Add to the general stats table
    headers = {
        "TOTAL_READS": {
            "title": f"{config.read_count_prefix} Reads",
            "description": f"The total number of reads in the input file ({config.read_count_desc})",
            "modify": lambda x: x * config.read_count_multiplier,
            "scale": "Blues",
            "shared_key": "read_count",
        }
    }
    module.general_stats_addcols(data_by_sample, headers, namespace="QualityYieldMetrics")

    return len(data_by_sample)



""" MultiQC module to parse base misincorporation output from mapdamage2 """


import logging
import os

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import linegraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """mapDamage module"""

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="mapDamage",
            anchor="mapdamage",
            href="https://github.com/ginolhac/mapDamage",
            info="mapDamage: tracking and quantifying damage patterns in ancient DNA sequences.",
            doi="https://doi.org/10.1093/bioinformatics/btt193",
        )

        # Init empty dictionaries
        self.threepGtoAfreq_data = dict()
        self.fivepCtoTfreq_data = dict()
        self.lgdist_fw_data = dict()
        self.lgdist_rv_data = dict()

        # Find and load log files
        for f in self.find_log_files("mapdamage", filehandles=True):
            self.parse_logs(f)

        # Filter to strip out ignored sample names
        self.threepGtoAfreq_data = self.ignore_samples(self.threepGtoAfreq_data)
        self.fivepCtoTfreq_data = self.ignore_samples(self.fivepCtoTfreq_data)
        self.lgdist_fw_data = self.ignore_samples(self.lgdist_fw_data)
        self.lgdist_rv_data = self.ignore_samples(self.lgdist_rv_data)

        ## Stop computations if there is no data after ignoring samples
        if (
            len(self.threepGtoAfreq_data) == 0
            and len(self.fivepCtoTfreq_data) == 0
            and len(self.lgdist_fw_data) == 0
            and len(self.lgdist_rv_data) == 0
        ):
            raise ModuleNoSamplesFound

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        ## No need to run self.write_data_file as all the data imported is already in a TSV format and can be (almost) directly used for plotting.

        # Basic Stats Table, use generic function to add data to general table
        self.dmgprof_misinc_stats(self.threepGtoAfreq_data, "3 Prime", "G>A")
        self.dmgprof_misinc_stats(self.fivepCtoTfreq_data, "5 Prime", "C>T")

        # Add plots
        if len(self.threepGtoAfreq_data) > 0:
            self.add_section(
                name="3P misincorporation plot",
                description="3' misincorporation plot for G>A substitutions",
                helptext="""
                This plot shows the frequency of G>A substitutions at the 3' read ends. Typically, one would observe high substitution percentages for ancient DNA, whereas modern DNA does not show these in higher extents.
                """,
                plot=self.threeprime_plot(),
            )
        if len(self.fivepCtoTfreq_data) > 0:
            self.add_section(
                name="5P misincorporation plot",
                description="5' misincorporation plot for C>T substitutions",
                helptext="""
                This plot shows the frequency of C>T substitutions at the 5' read ends. Typically, one would observe high substitution percentages for ancient DNA, whereas modern DNA does not show these in higher extents.
                """,
                plot=self.fiveprime_plot(),
            )
        if len(self.lgdist_fw_data) > 0 and len(self.lgdist_rv_data) > 0:
            self.add_section(
                name="Forward read length distribution",
                description="Read length distribution for forward strand (+) reads.",
                helptext="""
                This plot shows the read length distribution of the forward reads in the investigated sample. Reads below lengths of 30bp are typically filtered, so the plot doesn't show these in many cases. A shifted distribution of read lengths towards smaller read lengths (e.g around 30-50bp) is also an indicator of ancient DNA.
                """,
                plot=self.lgdistplot(self.lgdist_fw_data, "Forward"),
            )
            self.add_section(
                name="Reverse read length distribution",
                description="Read length distribution for reverse strand (−) reads.",
                helptext="""
                This plot shows the read length distribution of the reverse reads in the investigated sample. Reads below lengths of 30bp are typically filtered, so the plot doesn't show these in many cases. A shifted distribution of read lengths towards smaller read lengths (e.g around 30-50bp) is also an indicator of ancient DNA.
                """,
                plot=self.lgdistplot(self.lgdist_rv_data, "Reverse"),
            )

    ## Parse misincorporation file into a dict with 1 key ('dmg_5p' or 'dmg_3p') and 1 value (list of 2nd column values)
    def parse_misincorporation(self, f):
        """Parse the misincorporation data from mapDamage output files"""
        misincorporation_dict = {}
        values = []
        try:
            for line in f["f"]:
                if line.startswith("#"):
                    continue
                else:
                    line = line.rstrip().split("\t")
                    if line[0] == "pos":
                        readend = line[1][0:2]
                    else:
                        values.append(float(line[1]))
            misincorporation_dict["dmg_" + readend] = values
            return misincorporation_dict
        except Exception as e:
            print(e)
            log.warning(f"Could not parse mapDamage misincorporation file: '{f['fn']}'")
            return None

    ## Parse length distribution file into a dict with 2 keys ('lendist_fw' and 'lendist_rv') and 1 value each (dict of {length : count})
    def parse_length_distribution(self, f):
        """Parse the length distribution data from mapDamage output files"""
        length_distribution_dict = {"lendist_fw": {}, "lendist_rv": {}}
        try:
            for line in f["f"]:
                ## Skip commented lines and header
                if line.startswith("#") or line.startswith("Std"):
                    continue
                else:
                    ## Lines are tab-separated, first column is strand, second column is read length, third column is count
                    line = line.rstrip().split("\t")
                    if line[0] == "+":
                        strand = "fw"
                    else:
                        strand = "rv"

                    length_distribution_dict["lendist_" + strand][line[1]] = int(line[2])
            return length_distribution_dict
        except Exception as e:
            print(e)
            log.warning(f"Could not parse mapDamage length distribution file: '{f['fn']}'")
            return None

    # Parse input files
    def parse_logs(self, f):
        """Parse the output from mapDamage"""

        # Get sample name from result directory name
        ## Remove the *results_ prefix from the sample name if it is there.
        s_name = self.clean_s_name(f["root"], f, root=os.path.dirname(f["root"])).lstrip("*results_")
        self.add_data_source(f, s_name)

        if f["fn"].endswith("_freq.txt") and f["fn"].startswith("3p"):
            # Add 3' G to A data
            self.threepGtoAfreq_data[s_name] = self.parse_misincorporation(f)["dmg_3p"]

        elif f["fn"].endswith("_freq.txt") and f["fn"].startswith("5p"):
            # Add 5' C to T data
            self.fivepCtoTfreq_data[s_name] = self.parse_misincorporation(f)["dmg_5p"]

        elif f["fn"] == "lgdistribution.txt":
            # Add lendist forward & reverse
            lgdist_data = self.parse_length_distribution(f)
            self.lgdist_fw_data[s_name] = lgdist_data["lendist_fw"]
            self.lgdist_rv_data[s_name] = lgdist_data["lendist_rv"]

    #### Tables from here on

    def dmgprof_misinc_stats(self, dict_to_plot, readend, substitution):
        """Take the parsed stats from the mapDamage and add it to the
        basic stats table at the top of the report"""

        headers = {
            f"mapdamage-{readend}1": {
                "id": f"misinc-stats-1st-{readend}-{substitution}",
                "title": f"{readend} {substitution} 1st base",
                "description": f"{readend} 1st base substitution frequency for {substitution}",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "YlGnBu",
                "modify": lambda x: x * 100.0,
            },
            f"mapdamage-{readend}2": {
                "id": f"misinc-stats-2nd-{readend}-{substitution}",
                "title": f"{readend} {substitution} 2nd base",
                "description": f"{readend} 2nd base substitution frequency for {substitution}",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "BuGn",
                "hidden": True,
                "modify": lambda x: x * 100.0,
            },
        }

        # Create new small subset dictionary for entries (we need just the first two data (k,v) pairs from each report)
        # Only the first two parts are informative from both 3' and 5' ends of reads (1st, 2nd base damage pattern)
        dict_to_add = dict()

        for key in dict_to_plot.keys():
            tmp = dict_to_plot[key]
            pos = ["mapdamage-" + readend + "1", "mapdamage-" + readend + "2"]
            strlist = tmp[:2]
            tuples = list(zip(pos, strlist))
            data = dict((x, y) for x, y in tuples)
            # Extract first two elements from list
            dict_to_add[key] = data

        self.general_stats_addcols(dict_to_add, headers)

    #### Plotting from here on
    # Nice Linegraph plot for lgdist data

    def lgdistplot(self, dict_to_use, orientation):
        """Generate a read length distribution plot"""

        data = dict()
        for s_name in dict_to_use:
            try:
                data[s_name] = {int(d): int(dict_to_use[s_name][d]) for d in dict_to_use[s_name]}
            except KeyError:
                pass
        if len(data) == 0:
            log.debug("No valid data for forward read lgdist input!")
            return None

        config = {
            "id": f"mapdamage-length-distribution-{orientation}",
            "title": f"mapDamage: Read length distribution - {orientation} ",
            "ylab": "Number of reads",
            "xlab": "Readlength (bp)",
            "xDecimals": False,
            "tt_label": "{point.y} reads of length {point.x}",
            "ymin": 0,
            "xmin": 0,
        }
        return linegraph.plot(data, config)

    # Linegraph plot for 3pGtoA
    def threeprime_plot(self):
        """Generate a 3' G>A linegraph plot"""

        data = dict()
        dict_to_add = dict()
        # Create tuples out of entries
        for key in self.threepGtoAfreq_data:
            pos = list(range(1, len(self.threepGtoAfreq_data.get(key))))
            # Multiply values by 100 to get %
            tmp = [i * 100.0 for i in self.threepGtoAfreq_data.get(key)]
            tuples = list(zip(pos, tmp))
            # Get a dictionary out of it
            data = dict((x, y) for x, y in tuples)
            dict_to_add[key] = data

        config = {
            "id": "mapdamage-threeprime_misinc_plot",
            "title": "mapDamage: 3' G>A misincorporation plot",
            "ylab": "% G to A substituted",
            "xlab": "Nucleotide position from 3'",
            "tt_label": "{point.y:.2f} % G>A misincorporations at nucleotide position {point.x}",
            "ymin": 0,
            "xmin": 1,
        }

        return linegraph.plot(dict_to_add, config)

    # Linegraph plot for 5pCtoT
    def fiveprime_plot(self):
        """Generate a 5' C>T linegraph plot"""

        data = dict()
        dict_to_add = dict()
        # Create tuples out of entries
        for key in self.fivepCtoTfreq_data:
            pos = list(range(1, len(self.fivepCtoTfreq_data.get(key))))
            tmp = [i * 100.0 for i in self.fivepCtoTfreq_data.get(key)]
            tuples = list(zip(pos, tmp))
            # Get a dictionary out of it
            data = dict((x, y) for x, y in tuples)
            dict_to_add[key] = data

        config = {
            "id": "mapdamage-fiveprime_misinc_plot",
            "title": "mapDamage: 5' C>T misincorporation plot",
            "ylab": "% C to T substituted",
            "xlab": "Nucleotide position from 5'",
            "tt_label": "{point.y:.2f} % C>T misincorporations at nucleotide position {point.x}",
            "ymin": 0,
            "xmin": 1,
        }

        return linegraph.plot(dict_to_add, config)



""" MultiQC submodule to parse output from Picard WgsMetrics """

import logging

from multiqc import config
from multiqc.modules.picard import util
from multiqc.plots import bargraph, linegraph

# Initialise the logger
log = logging.getLogger(__name__)


def parse_reports(module):
    """Find Picard WgsMetrics reports and parse their data"""

    # Set up vars
    data_by_sample = dict()
    histogram_by_sample = dict()

    picard_config = getattr(config, "picard_config", {})
    skip_histo = picard_config.get("wgsmetrics_skip_histogram", False)

    # Go through logs and find Metrics
    for f in module.find_log_files("picard/wgs_metrics", filehandles=True):
        # Sample name from input file name by default
        s_name = f["s_name"]
        in_hist = False

        for line in f["f"]:
            maybe_s_name = util.extract_sample_name(
                module,
                line,
                f,
                picard_tool="CollectWgsMetrics",
            )
            if maybe_s_name:
                s_name = maybe_s_name

            # Catch the histogram values
            if in_hist and not skip_histo:
                try:
                    sections = line.split("\t")
                    cov = int(sections[0])
                    count = int(sections[1])
                    histogram_by_sample[s_name][cov] = count
                except ValueError:
                    # Reset in case we have more in this log file
                    s_name = None
                    in_hist = False

            if s_name is None:
                continue

            if util.is_line_right_before_table(line, picard_class="WgsMetrics"):
                keys = f["f"].readline().strip("\n").split("\t")
                vals = f["f"].readline().strip("\n").split("\t")
                if len(vals) != len(keys):
                    continue

                if s_name in data_by_sample:
                    log.debug(f"Duplicate sample name found in {f['fn']}! Overwriting: {s_name}")

                module.add_data_source(f, s_name, section="WgsMetrics")
                data_by_sample[s_name] = dict()

                for k, v in zip(keys, vals):
                    try:
                        v = float(v)
                    except ValueError:
                        pass
                    data_by_sample[s_name][k] = v

            elif line.startswith("## HISTOGRAM"):
                keys = f["f"].readline().strip("\n").split("\t")
                assert len(keys) >= 2, (keys, f)
                in_hist = True
                histogram_by_sample[s_name] = dict()

    # Filter to strip out ignored sample names
    data_by_sample = module.ignore_samples(data_by_sample)
    if len(data_by_sample) == 0:
        return 0

    # Superfluous function call to confirm that it is used in this module
    # Replace None with actual version if it is available
    module.add_software_version(None)

    # Write parsed data to a file
    module.write_data_file(data_by_sample, "multiqc_picard_wgsmetrics")

    # Add to general stats table
    headers = dict()
    headers["MEDIAN_COVERAGE"] = {
        "title": "Median Coverage",
        "description": "The median coverage in bases of the genome territory, after all filters are applied.",
        "min": 0,
        "suffix": "X",
        "scale": "GnBu",
    }
    headers["MEAN_COVERAGE"] = {
        "title": "Mean Coverage",
        "description": "The mean coverage in bases of the genome territory, after all filters are applied.",
        "min": 0,
        "suffix": "X",
        "scale": "GnBu",
        "hidden": True,
    }
    headers["SD_COVERAGE"] = {
        "title": "SD Coverage",
        "description": "The standard deviation coverage in bases of the genome territory, after all filters are applied.",
        "min": 0,
        "suffix": "X",
        "scale": "GnBu",
        "hidden": True,
    }
    # user configurable coverage level
    try:
        covs = config.picard_config["general_stats_target_coverage"]
        assert isinstance(covs, list)
        assert len(covs) > 0
        covs = [str(i) for i in covs]
        log.debug(f"Custom Picard coverage thresholds: {', '.join([i for i in covs])}")
    except (AttributeError, TypeError, AssertionError, KeyError):
        covs = ["30"]
    for c in covs:
        headers[f"PCT_{c}X"] = {
            "id": f"picard_target_bases_{c}X",
            "title": f"Bases ≥ {c}X",
            "description": f"Percent of target bases with coverage ≥ {c}X",
            "max": 100,
            "min": 0,
            "suffix": "%",
            "format": "{:,.0f}",
            "scale": "RdYlGn",
            "modify": lambda x: util.multiply_hundred(x),
        }
    module.general_stats_addcols(data_by_sample, headers, namespace="WgsMetrics")

    # Section with histogram plot
    if histogram_by_sample and not skip_histo:
        # Figure out where to cut histogram tail
        max_cov = picard_config.get("wgsmetrics_histogram_max_cov")
        if max_cov is None:
            max_cov = 10
            for s_name, hist in histogram_by_sample.items():
                total = float(sum(hist.values()))
                running_total = 0
                for k, v in hist.items():
                    running_total += v
                    if running_total > total * 0.99:
                        max_cov = max(k, max_cov)
                        break

        # Cut histogram tail and make a normalised percentage version of the data plus dropoff
        data = {}
        data_percent = {}
        maxval = 0
        for s_name, hist in histogram_by_sample.items():
            data[s_name] = dict()
            data_percent[s_name] = dict()
            total = float(sum(hist.values()))
            cumulative = 0
            for k, v in hist.items():
                if k <= max_cov:
                    cumulative += v
                    data[s_name][k] = v
                    maxval = max(maxval, v)
                    data_percent[s_name][k] = 100 - (cumulative / total) * 100
                else:
                    break

        # Plot the histogram data and add section
        pconfig = {
            "id": "picard_wgs_metrics_histogram",
            "title": "Picard: WGS Coverage",
            "ylab": "Percentage of Bases",
            "xlab": "Fold Coverage",
            "xDecimals": False,
            "tt_label": "<b>{point.x} X</b>: {point.y:.1f}",
            "ymin": 0,
            "ymax": 100,
            "smooth_points": picard_config.get("wgsmetrics_histogram_smooth", 1000),
            "data_labels": [
                {"name": "Percentage Drop-Off", "ylab": "Percentage of Bases", "ymax": 100},
                {"name": "Counts Histogram", "ylab": "Coverage", "ymax": maxval},
            ],
        }
        module.add_section(
            name="WGS Coverage",
            anchor="picard-wgsmetrics-cov",
            description="The number of bases in the genome territory for each fold coverage. "
            + "Note that final 1% of data is hidden to prevent very long tails.",
            plot=linegraph.plot([data_percent, data], pconfig),
        )

    # Bar plot of ignored bases
    pdata = dict()
    for s_name, data in data_by_sample.items():
        pdata[s_name] = dict()
        pdata[s_name]["PCT_EXC_MAPQ"] = data["PCT_EXC_MAPQ"] * 100.0
        pdata[s_name]["PCT_EXC_DUPE"] = data["PCT_EXC_DUPE"] * 100.0
        pdata[s_name]["PCT_EXC_UNPAIRED"] = data["PCT_EXC_UNPAIRED"] * 100.0
        pdata[s_name]["PCT_EXC_BASEQ"] = data["PCT_EXC_BASEQ"] * 100.0
        pdata[s_name]["PCT_EXC_OVERLAP"] = data["PCT_EXC_OVERLAP"] * 100.0
        pdata[s_name]["PCT_EXC_CAPPED"] = data["PCT_EXC_CAPPED"] * 100.0

    keys = dict()
    keys["PCT_EXC_MAPQ"] = {"name": "Low mapping quality"}
    keys["PCT_EXC_DUPE"] = {"name": "Duplicates reads"}
    keys["PCT_EXC_UNPAIRED"] = {"name": "No mapped mate pair"}
    keys["PCT_EXC_BASEQ"] = {"name": "Low base quality"}
    keys["PCT_EXC_OVERLAP"] = {"name": "Overlapping insert"}
    keys["PCT_EXC_CAPPED"] = {"name": "Over capped coverage"}

    # Config for the plot
    pconfig = {
        "id": "picard_wgs_metrics_bases",
        "title": "Picard: WGS Filtered Bases",
        "cpswitch": False,
        "ylab": "% Bases",
        "ymax": 100,
    }

    module.add_section(
        name="WGS Filtered Bases",
        anchor="picard-wgsmetrics-bases",
        description="For more information about the filtered categories, see the "
        + '<a href="http://broadinstitute.github.io/picard/picard-metric-definitions.html#CollectWgsMetrics.WgsMetrics" target="_blank">Picard documentation</a>.',
        plot=bargraph.plot(pdata, keys, pconfig),
    )

    # Return the number of detected samples to the parent module
    return len(data_by_sample)



""" MultiQC submodule to parse output from Picard MarkIlluminaAdapters """

import logging

from multiqc.plots import linegraph

from .util import read_histogram

# Initialise the logger
log = logging.getLogger(__name__)


def parse_reports(self):
    """Find Picard MarkIlluminaAdapters reports and parse their data"""

    headers = ["clipped_bases", "read_count"]
    formats = [int, int]
    data_by_sample = read_histogram(
        self,
        program_key="picard/markilluminaadapters",
        headers=headers,
        formats=formats,
        picard_tool="MarkIlluminaAdapters",
    )

    # Filter to strip out ignored sample names
    data_by_sample = self.ignore_samples(data_by_sample)
    if not data_by_sample:
        return 0

    # Superfluous function call to confirm that it is used in this module
    # Replace None with actual version if it is available
    self.add_software_version(None)

    # Write parsed data to a file
    self.write_data_file(data_by_sample, "multiqc_picard_mark_illumina_adapters")

    # Plot the data and add section
    pconfig = {
        "id": "picard_mark_illumina_adapters",
        "title": "Picard: Mark Illumina Adapters",
        "ylab": "Clipped Bases",
        "xlab": "Cycle Number",
        "xDecimals": False,
        "tt_label": "<b>Cycle {point.x}</b>: {point.y:.2f}",
        "ymin": 0,
    }

    lg = {}
    for s_name in data_by_sample:
        lg[s_name] = {clipped_bases: data["read_count"] for clipped_bases, data in data_by_sample[s_name].items()}

    self.add_section(
        name="Mark Illumina Adapters",
        description="""
            Number of Clipped Bases by Read.
            See the [Picard Docuementation](https://broadinstitute.github.io/picard/command-line-overview.html#MarkIlluminaAdapters) for details.
        """,
        anchor="picard-mark-illumina-adapters",
        plot=linegraph.plot(lg, pconfig),
    )

    # Return the number of detected samples to the parent module
    return len(data_by_sample)



""" MultiQC submodule to parse output from deepTools bamPEFragmentSize for read length distribution """

import logging

from multiqc.plots import linegraph

# Initialise the logger
log = logging.getLogger(__name__)


class bamPEFragmentSizeDistributionMixin:
    def parse_bamPEFragmentSizeDistribution(self):
        """Find bamPEFragmentSize output. Supports the --outRawFragmentLengths option"""
        self.deeptools_bamPEFragmentSizeDistribution = dict()
        for f in self.find_log_files("deeptools/bamPEFragmentSizeDistribution", filehandles=False):
            parsed_data = self.parseBamPEFDistributionFile(f)
            for k, v in parsed_data.items():
                if k in self.deeptools_bamPEFragmentSizeDistribution:
                    log.warning(f"Replacing duplicate sample {k}.")
                self.deeptools_bamPEFragmentSizeDistribution[k] = v
            if len(parsed_data) > 0:
                self.add_data_source(f, section="bamPEFragmentSizeDistribution")

            # Superfluous function call to confirm that it is used in this module
            # Replace None with actual version if it is available
            self.add_software_version(None, f["s_name"])

        self.deeptools_bamPEFragmentSizeDistribution = self.ignore_samples(self.deeptools_bamPEFragmentSizeDistribution)

        if len(self.deeptools_bamPEFragmentSizeDistribution) > 0:
            # Write data to file
            self.write_data_file(self.deeptools_bamPEFragmentSizeDistribution, "deeptools_frag_size_dist")

            config = {
                "id": "fragment_size_distribution_plot",
                "title": "deeptools: Fragment Size Distribution Plot",
                "ylab": "Occurrence",
                "xlab": "Fragment Size (bp)",
                "smooth_points": 50,
                "xmax": 1000,
                "xDecimals": False,
                "tt_label": "<b>Fragment Size (bp) {point.x}</b>: {point.y} Occurrence",
            }

            self.add_section(
                name="Fragment size distribution",
                anchor="fragment_size_distribution",
                description="Distribution of paired-end fragment sizes",
                plot=linegraph.plot(self.deeptools_bamPEFragmentSizeDistribution, config),
            )

        return len(self.deeptools_bamPEFragmentSizeDistribution)

    def parseBamPEFDistributionFile(self, f):
        d = dict()
        lastsample = []
        for line in f["f"].splitlines():
            cols = line.rstrip().split("\t")
            if cols[0] == "#bamPEFragmentSize":
                continue
            elif cols[0] == "Size":
                continue
            else:
                s_name = self.clean_s_name(cols[2].rstrip().split("/")[-1], f)
                if s_name != lastsample:
                    d[s_name] = dict()
                    lastsample = s_name
                d[s_name].update({self._int(cols[0]): self._int(cols[1])})

        return d



""" MultiQC functions to plot a violin plot """

import logging
from typing import List, Dict, Optional, Union

from multiqc.plots import table_object
from multiqc.utils import config
from multiqc.plots.plotly import violin

logger = logging.getLogger(__name__)

letters = "abcdefghijklmnopqrstuvwxyz"


# Load the template so that we can access its configuration
# Do this lazily to mitigate import-spaghetti when running unit tests
_template_mod = None


def get_template_mod():
    global _template_mod
    if not _template_mod:
        _template_mod = config.avail_templates[config.template].load()
    return _template_mod


def plot(data: List[Dict], headers: Optional[Union[List[Dict], Dict]] = None, pconfig=None):
    """Helper HTML for a violin plot.
    :param data: A list of data dicts
    :param headers: A list of dicts with information
                    for the series, such as colour scales, min and
                    max values etc.
    :return: HTML string
    """
    # Make a datatable object
    dt = table_object.DataTable(data, headers, pconfig)

    mod = get_template_mod()
    if "violin" in mod.__dict__ and callable(mod.violin):
        # noinspection PyBroadException
        try:
            return mod.violin(dt)
        except:  # noqa: E722
            if config.strict:
                # Crash quickly in the strict mode. This can be helpful for interactive
                # debugging of modules
                raise

    return violin.plot(dt)



""" MultiQC module to parse output from Bismark """


import logging
import re

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, beeswarm, linegraph

# Initialise the logger
log = logging.getLogger(__name__)

# Log parsing regexes
regexes = {
    "alignment": {
        "total_reads": r"Sequence(?:s| pairs) analysed in total:\s+(\d+)",
        "aligned_reads": r"Number of(?: paired-end)? alignments with a unique best hit(?: from the different alignments)?:\s+(\d+)",
        "no_alignments": r"Sequence(?:s| pairs) with no alignments under any condition:\s+(\d+)",
        "ambig_reads": r"Sequence(?:s| pairs) did not map uniquely:\s+(\d+)",
        "discarded_reads": r"Sequence(?:s| pairs) which were discarded because genomic sequence could not be extracted:\s+(\d+)",
        "total_c": r"Total number of C's analysed:\s+(\d+)",
        "meth_cpg": r"Total methylated C's in CpG context:\s+(\d+)",
        "meth_chg": r"Total methylated C's in CHG context:\s+(\d+)",
        "meth_chh": r"Total methylated C's in CHH context:\s+(\d+)",
        "unmeth_cpg": r"Total unmethylated C's in CpG context:\s+(\d+)",
        "unmeth_chg": r"Total unmethylated C's in CHG context:\s+(\d+)",
        "unmeth_chh": r"Total unmethylated C's in CHH context:\s+(\d+)",
        "percent_cpg_meth": r"C methylated in CpG context:\s+([\d\.]+)%",
        "percent_chg_meth": r"C methylated in CHG context:\s+([\d\.]+)%",
        "percent_chh_meth": r"C methylated in CHH context:\s+([\d\.]+)%",
        "strand_ot": r"CT(?:\/GA)?\/CT:\s+(\d+)\s+\(\(converted\) top strand\)",
        "strand_ctot": r"GA(?:\/CT)?\/CT:\s+(\d+)\s+\(complementary to \(converted\) top strand\)",
        "strand_ctob": r"GA(?:\/CT)?\/GA:\s+(\d+)\s+\(complementary to \(converted\) bottom strand\)",
        "strand_ob": r"CT(?:\/GA)?\/GA:\s+(\d+)\s+\(\(converted\) bottom strand\)",
        "strand_directional": r"Option '--(directional)' specified \(default mode\): alignments to complementary strands \(CTOT, CTOB\) were ignored \(i.e. not performed\)",
        "version": r"Bismark report for: .+ \(version: v([\d\.]+)\)",
    },
    "dedup": {
        "aligned_reads": r"Total number of alignments analysed in .+:\s+(\d+)",
        "dup_reads": r"Total number duplicated alignments removed:\s+(\d+)",
        "dup_reads_percent": r"Total number duplicated alignments removed:\s+\d+\s+\(([\d\.]+)%\)",
        "dedup_reads": r"Total count of deduplicated leftover sequences:\s+(\d+)",
        "dedup_reads_percent": r"Total count of deduplicated leftover sequences:\s+\d+\s+\(([\d\.]+)% of total\)",
    },
    "methextract": {
        "total_c": r"Total number of C's analysed:\s+(\d+)",
        "meth_cpg": r"Total methylated C's in CpG context:\s+(\d+)",
        "meth_chg": r"Total methylated C's in CHG context:\s+(\d+)",
        "meth_chh": r"Total methylated C's in CHH context:\s+(\d+)",
        "unmeth_cpg": r"Total C to T conversions in CpG context:\s+(\d+)",
        "unmeth_chg": r"Total C to T conversions in CHG context:\s+(\d+)",
        "unmeth_chh": r"Total C to T conversions in CHH context:\s+(\d+)",
        "percent_cpg_meth": r"C methylated in CpG context:\s+([\d\.]+)%",
        "percent_chg_meth": r"C methylated in CHG context:\s+([\d\.]+)%",
        "percent_chh_meth": r"C methylated in CHH context:\s+([\d\.]+)%",
        "version": r"Bismark Extractor Version: v([\d\.]+)",
    },
}


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Bismark",
            anchor="bismark",
            href="http://www.bioinformatics.babraham.ac.uk/projects/bismark/",
            info="is a tool to map bisulfite converted sequence reads and determine cytosine methylation states.",
            doi="10.1093/bioinformatics/btr167",
        )

        # Set up data structures
        self.bismark_data = {"alignment": {}, "dedup": {}, "methextract": {}, "bam2nuc": {}}
        self.bismark_mbias_data = {
            "meth": {"CpG_R1": {}, "CHG_R1": {}, "CHH_R1": {}, "CpG_R2": {}, "CHG_R2": {}, "CHH_R2": {}},
            "cov": {"CpG_R1": {}, "CHG_R1": {}, "CHH_R1": {}, "CpG_R2": {}, "CHG_R2": {}, "CHH_R2": {}},
        }

        # Find and parse bismark alignment reports
        for f in self.find_log_files("bismark/align"):
            parsed_data = self.parse_bismark_report(f["f"], regexes["alignment"])
            if parsed_data is not None:
                if "version" in parsed_data:
                    self.add_software_version(parsed_data["version"], f["s_name"])

                # Calculate percent_aligned - doubles as a good check that stuff has worked
                try:
                    parsed_data["percent_aligned"] = (parsed_data["aligned_reads"] / parsed_data["total_reads"]) * 100
                except (KeyError, ZeroDivisionError):
                    log.warning(f"Error calculating percentage for {f['fn']} - ignoring sample.")
                else:
                    if f["s_name"] in self.bismark_data["alignment"]:
                        log.debug(f"Duplicate alignment sample log found! Overwriting: {f['s_name']}")
                    self.add_data_source(f, section="alignment")
                    self.bismark_data["alignment"][f["s_name"]] = parsed_data

        # Find and parse bismark deduplication reports
        for f in self.find_log_files("bismark/dedup"):
            parsed_data = self.parse_bismark_report(f["f"], regexes["dedup"])
            if parsed_data is not None:
                if f["s_name"] in self.bismark_data["dedup"]:
                    log.debug(f"Duplicate deduplication sample log found! Overwriting: {f['s_name']}")
                self.add_data_source(f, section="deduplication")
                self.bismark_data["dedup"][f["s_name"]] = parsed_data

        # Find and parse bismark methylation extractor reports
        for f in self.find_log_files("bismark/meth_extract"):
            parsed_data = self.parse_bismark_report(f["f"], regexes["methextract"])
            s_name = f["s_name"]
            if parsed_data is not None:
                if s_name in self.bismark_data["methextract"]:
                    log.debug(f"Duplicate methylation extraction sample log found! Overwriting: {s_name}")

                if "version" in parsed_data:
                    self.add_software_version(parsed_data["version"], s_name)

                self.add_data_source(f, s_name, section="methylation_extraction")
                self.bismark_data["methextract"][s_name] = parsed_data

        # Find and parse M-bias plot data
        for f in self.find_log_files("bismark/m_bias", filehandles=True):
            self.parse_bismark_mbias(f)
            self.add_data_source(f, section="m_bias")

        # Find and parse bam2nuc reports
        for f in self.find_log_files("bismark/bam2nuc", filehandles=True):
            self.parse_bismark_bam2nuc(f)
            self.add_data_source(f, section="bam2nuc")

        # Filters to strip out ignored sample names
        for k in self.bismark_data:
            self.bismark_data[k] = self.ignore_samples(self.bismark_data[k])
        for k in self.bismark_mbias_data["meth"]:
            self.bismark_mbias_data["meth"][k] = self.ignore_samples(self.bismark_mbias_data["meth"][k])
        for k in self.bismark_mbias_data["cov"]:
            self.bismark_mbias_data["cov"][k] = self.ignore_samples(self.bismark_mbias_data["cov"][k])

        num_parsed = len(self.bismark_data["alignment"])
        num_parsed += len(self.bismark_data["dedup"])
        num_parsed += len(self.bismark_data["methextract"])
        num_parsed += len(self.bismark_mbias_data["meth"]["CpG_R1"])
        num_parsed += len(self.bismark_data["bam2nuc"])
        if num_parsed == 0:
            raise ModuleNoSamplesFound

        # Basic Stats Table
        self.bismark_stats_table()

        # Write out to the report
        if len(self.bismark_data["alignment"]) > 0:
            self.write_data_file(self.bismark_data["alignment"], "multiqc_bismark_alignment", sort_cols=True)
            log.info(f"Found {len(self.bismark_data['alignment'])} alignment reports")
            self.bismark_alignment_chart()

        if len(self.bismark_data["dedup"]) > 0:
            self.write_data_file(self.bismark_data["dedup"], "multiqc_bismark_dedup", sort_cols=True)
            log.info(f"Found {len(self.bismark_data['dedup'])} dedup reports")
            self.bismark_dedup_chart()

        if len(self.bismark_data["alignment"]) > 0:
            self.bismark_strand_chart()

        if len(self.bismark_data["methextract"]) > 0:
            self.write_data_file(self.bismark_data["methextract"], "multiqc_bismark_methextract", sort_cols=True)
            log.info(f"Found {len(self.bismark_data['methextract'])} methextract reports")
            self.bismark_methlyation_chart()

        if len(self.bismark_mbias_data["meth"]["CpG_R1"]) > 0:
            self.bismark_mbias_plot()

        if len(self.bismark_data["bam2nuc"]) > 0:
            self.write_data_file(self.bismark_data["bam2nuc"], "multiqc_bismark_bam2nuc", sort_cols=True)
            log.info(f"Found {len(self.bismark_data['bam2nuc'])} bismark bam2nuc reports")

    def parse_bismark_report(self, report, regexes):
        """Search a bismark report with a set of regexes"""
        parsed_data = {}
        for k, r in regexes.items():
            r_search = re.search(r, report, re.MULTILINE)
            if r_search:
                try:
                    parsed_data[k] = float(r_search.group(1))
                except ValueError:
                    parsed_data[k] = r_search.group(1)  # NaN
        if len(parsed_data) == 0:
            return None
        return parsed_data

    def parse_bismark_mbias(self, f):
        """Parse the Bismark M-Bias plot data"""
        s = f["s_name"]
        self.bismark_mbias_data["meth"]["CpG_R1"][s] = {}
        self.bismark_mbias_data["meth"]["CHG_R1"][s] = {}
        self.bismark_mbias_data["meth"]["CHH_R1"][s] = {}
        self.bismark_mbias_data["cov"]["CpG_R1"][s] = {}
        self.bismark_mbias_data["cov"]["CHG_R1"][s] = {}
        self.bismark_mbias_data["cov"]["CHH_R1"][s] = {}
        self.bismark_mbias_data["meth"]["CpG_R2"][s] = {}
        self.bismark_mbias_data["meth"]["CHG_R2"][s] = {}
        self.bismark_mbias_data["meth"]["CHH_R2"][s] = {}
        self.bismark_mbias_data["cov"]["CpG_R2"][s] = {}
        self.bismark_mbias_data["cov"]["CHG_R2"][s] = {}
        self.bismark_mbias_data["cov"]["CHH_R2"][s] = {}
        key = None
        for line in f["f"]:
            if "context" in line:
                if "CpG" in line:
                    key = "CpG"
                elif "CHG" in line:
                    key = "CHG"
                elif "CHH" in line:
                    key = "CHH"
                if "(R1)" in line:
                    key += "_R1"
                elif "(R2)" in line:
                    key += "_R2"
                else:
                    key += "_R1"
            if key is not None:
                sections = line.split()
                try:
                    pos = int(sections[0])
                    self.bismark_mbias_data["meth"][key][s][pos] = float(sections[3])
                    self.bismark_mbias_data["cov"][key][s][pos] = int(sections[4])
                except (IndexError, ValueError):
                    continue

        # Remove empty dicts (e.g. R2 for SE data)
        for t in self.bismark_mbias_data:
            for k in self.bismark_mbias_data[t]:
                self.bismark_mbias_data[t][k] = {
                    s_name: self.bismark_mbias_data[t][k][s_name]
                    for s_name in self.bismark_mbias_data[t][k]
                    if len(self.bismark_mbias_data[t][k][s_name]) > 0
                }

    def parse_bismark_bam2nuc(self, f):
        """Parse reports generated by Bismark bam2nuc"""
        if f["s_name"] in self.bismark_data["bam2nuc"]:
            log.debug(f"Duplicate deduplication sample log found! Overwriting: {f['s_name']}")
        self.add_data_source(f, section="bam2nuc")
        self.bismark_data["bam2nuc"][f["s_name"]] = dict()

        headers = None
        for line in f["f"]:
            sections = line.rstrip().split("\t")
            if headers is None:
                headers = sections
            else:
                k = None
                for i, h in enumerate(headers):
                    if i == 0:
                        k = sections[0]
                    elif k is not None:
                        key = f"{k}_{h.lower().replace(' ', '_')}"
                        self.bismark_data["bam2nuc"][f["s_name"]][key] = sections[i]

    def bismark_stats_table(self):
        """Take the parsed stats from the Bismark reports and add them to the
        basic stats table at the top of the report"""

        headers = {
            "alignment": dict(),
            "dedup": dict(),
            "methextract": dict(),
            "bam2nuc": dict(),
        }
        headers["methextract"]["percent_cpg_meth"] = {
            "title": "% mCpG",
            "description": "% Cytosines methylated in CpG context",
            "max": 100,
            "min": 0,
            "suffix": "%",
            "scale": "Greens",
        }
        headers["methextract"]["percent_chg_meth"] = {
            "title": "% mCHG",
            "description": "% Cytosines methylated in CHG context",
            "max": 100,
            "min": 0,
            "suffix": "%",
            "scale": "Oranges",
        }
        headers["methextract"]["percent_chh_meth"] = {
            "title": "% mCHH",
            "description": "% Cytosines methylated in CHH context",
            "max": 100,
            "min": 0,
            "suffix": "%",
            "scale": "Oranges",
        }
        headers["methextract"]["total_c"] = {
            "title": "M C's",
            "description": "Total number of C's analysed, in millions",
            "min": 0,
            "scale": "Purples",
            "modify": lambda x: x / 1000000,
        }
        headers["bam2nuc"]["C_coverage"] = {
            "title": "C Coverage",
            "description": "Cyotosine Coverage",
            "min": 0,
            "suffix": "X",
            "scale": "Greens",
            "format": "{:,.2f}",
        }
        headers["dedup"]["dup_reads_percent"] = {
            "title": "% Dups",
            "description": "Percent Duplicated Alignments",
            "max": 100,
            "min": 0,
            "suffix": "%",
            "scale": "RdYlGn-rev",
        }
        headers["dedup"]["dedup_reads"] = {
            "title": f"{config.read_count_prefix} Unique",
            "description": f"Deduplicated Alignments ({config.read_count_desc})",
            "min": 0,
            "scale": "Greens",
            "modify": lambda x: x * config.read_count_multiplier,
            "shared_key": "read_count",
            "hidden": True,
        }
        headers["alignment"]["aligned_reads"] = {
            "title": f"{config.read_count_prefix} Aligned",
            "description": f"Total Aligned Sequences ({config.read_count_desc})",
            "min": 0,
            "scale": "PuRd",
            "modify": lambda x: x * config.read_count_multiplier,
            "shared_key": "read_count",
            "hidden": True,
        }
        headers["alignment"]["percent_aligned"] = {
            "title": "% Aligned",
            "description": "Percent Aligned Sequences",
            "max": 100,
            "min": 0,
            "suffix": "%",
            "scale": "YlGn",
        }

        self.general_stats_addcols(self.bismark_data["methextract"], headers["methextract"])
        self.general_stats_addcols(self.bismark_data["bam2nuc"], headers["bam2nuc"])
        self.general_stats_addcols(self.bismark_data["dedup"], headers["dedup"])
        self.general_stats_addcols(self.bismark_data["alignment"], headers["alignment"])

    def bismark_alignment_chart(self):
        """Make the alignment plot"""

        # Specify the order of the different possible categories
        keys = {
            "aligned_reads": {"color": "#2f7ed8", "name": "Aligned Uniquely"},
            "ambig_reads": {"color": "#492970", "name": "Aligned Ambiguously"},
            "no_alignments": {"color": "#0d233a", "name": "Did Not Align"},
            "discarded_reads": {"color": "#f28f43", "name": "No Genomic Sequence"},
        }

        # Config for the plot
        config = {
            "id": "bismark_alignment",
            "title": "Bismark: Alignment Scores",
            "ylab": "# Reads",
            "cpswitch_counts_label": "Number of Reads",
        }

        self.add_section(
            name="Alignment Rates",
            anchor="bismark-alignment",
            plot=bargraph.plot(self.bismark_data["alignment"], keys, config),
        )

    def bismark_strand_chart(self):
        """Make the strand alignment plot"""

        # Specify the order of the different possible categories
        keys = {
            "strand_ob": {"name": "Original bottom strand"},
            "strand_ctob": {"name": "Complementary to original bottom strand"},
            "strand_ctot": {"name": "Complementary to original top strand"},
            "strand_ot": {"name": "Original top strand"},
        }

        # See if we have any directional samples
        directional = 0
        d_mode = ""
        for sn in self.bismark_data["alignment"].values():
            if "strand_directional" in sn.keys():
                directional += 1
        if directional == len(self.bismark_data["alignment"]):
            keys.pop("strand_ctob", None)
            keys.pop("strand_ctot", None)
            d_mode = "All samples were run with <code>--directional</code> mode; alignments to complementary strands (CTOT, CTOB) were ignored."
        elif directional > 0:
            d_mode = "{} samples were run with <code>--directional</code> mode; alignments to complementary strands (CTOT, CTOB) were ignored.".format(
                directional
            )

        # Config for the plot
        config = {
            "id": "bismark_strand_alignment",
            "title": "Bismark: Alignment to Individual Bisulfite Strands",
            "ylab": "% Reads",
            "cpswitch_c_active": False,
            "cpswitch_counts_label": "Number of Reads",
        }

        self.add_section(
            name="Strand Alignment",
            anchor="bismark-strands",
            description=d_mode,
            plot=bargraph.plot(self.bismark_data["alignment"], keys, config),
        )

    def bismark_dedup_chart(self):
        """Make the deduplication plot"""

        # Specify the order of the different possible categories
        keys = {
            "dedup_reads": {"name": "Deduplicated reads (remaining)"},
            "dup_reads": {"name": "Duplicate reads (removed)"},
        }

        # Config for the plot
        config = {
            "id": "bismark_deduplication",
            "title": "Bismark: Deduplication",
            "ylab": "% Reads",
            "cpswitch_c_active": False,
            "cpswitch_counts_label": "Number of Reads",
        }

        self.add_section(
            name="Deduplication",
            anchor="bismark-deduplication",
            plot=bargraph.plot(self.bismark_data["dedup"], keys, config),
        )

    def bismark_methlyation_chart(self):
        """Make the methylation plot"""

        # Config for the plot
        defaults = {"max": 100, "min": 0, "suffix": "%", "decimalPlaces": 1}
        keys = {
            "percent_cpg_meth": dict(defaults, **{"title": "Methylated CpG"}),
            "percent_chg_meth": dict(defaults, **{"title": "Methylated CHG"}),
            "percent_chh_meth": dict(defaults, **{"title": "Methylated CHH"}),
        }

        self.add_section(
            name="Cytosine Methylation",
            anchor="bismark-methylation",
            plot=beeswarm.plot(
                self.bismark_data["methextract"],
                keys,
                {
                    "id": "bismark-methylation-dp",
                    "title": "Bismark: Cytosine Methylation",
                },
            ),
        )

    def bismark_mbias_plot(self):
        """Make the M-Bias plot"""

        description = '<p>This plot shows the average percentage methylation and coverage across reads. See the \n\
        <a href="https://felixkrueger.github.io/Bismark/bismark/methylation_extraction/#m-bias-plot" target="_blank">bismark user guide</a> \n\
        for more information on how these numbers are generated.</p>'

        pconfig = {
            "id": "bismark_mbias",
            "title": "Bismark: M-Bias",
            "ylab": "% Methylation",
            "xlab": "Position (bp)",
            "xDecimals": False,
            "ymax": 100,
            "ymin": 0,
            "tt_label": "<b>{point.x} bp</b>: {point.y:.1f}%",
            "data_labels": [
                {"name": "CpG R1", "ylab": "% Methylation", "ymax": 100},
                {"name": "CHG R1", "ylab": "% Methylation", "ymax": 100},
                {"name": "CHH R1", "ylab": "% Methylation", "ymax": 100},
            ],
        }
        datasets = [
            self.bismark_mbias_data["meth"]["CpG_R1"],
            self.bismark_mbias_data["meth"]["CHG_R1"],
            self.bismark_mbias_data["meth"]["CHH_R1"],
        ]

        if len(self.bismark_mbias_data["meth"]["CpG_R2"]) > 0:
            pconfig["data_labels"].append({"name": "CpG R2", "ylab": "% Methylation", "ymax": 100})
            pconfig["data_labels"].append({"name": "CHG R2", "ylab": "% Methylation", "ymax": 100})
            pconfig["data_labels"].append({"name": "CHH R2", "ylab": "% Methylation", "ymax": 100})
            datasets.append(self.bismark_mbias_data["meth"]["CpG_R2"])
            datasets.append(self.bismark_mbias_data["meth"]["CHG_R2"])
            datasets.append(self.bismark_mbias_data["meth"]["CHH_R2"])

        self.add_section(
            name="M-Bias", anchor="bismark-mbias", description=description, plot=linegraph.plot(datasets, pconfig)
        )



#!/usr/bin/env python

""" MultiQC plugin hooks. Enables MultiQC plugins
to run their own custom subroutines at predefined
trigger points during MultiQC execution. """

from importlib_metadata import entry_points

# Load the hooks
hook_functions = {}
for entry_point in entry_points(group="multiqc.hooks.v1"):
    try:
        hook_functions[entry_point.name].append(entry_point.load())
    except KeyError:
        hook_functions[entry_point.name] = [entry_point.load()]


# Function to run the hooks
def mqc_trigger(trigger):
    for hook in hook_functions.get(trigger, []):
        hook()



""" MultiQC module to parse output from MetaPhlAn """


import logging
import re

from multiqc.utils import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """MetaPhlAn module"""

    def __init__(
        # Initialise the parent object
        self,
        name="MetaPhlAn",
        anchor="metaphlan",
        href="https://github.com/biobakery/MetaPhlAn",
        info="is a computational tool for profiling the composition of microbial communities from metagenomic shotgun sequencing data.",
        doi="10.1038/s41587-023-01688-w",
    ):
        super(MultiqcModule, self).__init__(
            name=name,
            anchor=anchor,
            href=href,
            info=info,
            doi=doi,
        )
        # Custom options from user config that can overwrite base module values
        self.t_ranks = {
            "s": "Species",
            "g": "Genus",
            "f": "Family",
            "o": "Order",
            "c": "Class",
            "p": "Phylum",
            "k": "Kingdom",
        }

        self.top_n = getattr(config, "metaphlan", {}).get("top_n", 10)

        self.metaphlan_raw_data = dict()
        for f in self.find_log_files("metaphlan", filehandles=True):
            f["f"].seek(0)
            self.parse_logs(f)
            self.add_data_source(f)

        self.metaphlan_raw_data = self.ignore_samples(self.metaphlan_raw_data)

        if len(self.metaphlan_raw_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.metaphlan_raw_data)} reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        self.write_data_file(self.metaphlan_raw_data, f"multiqc_{self.anchor}")

        # Sum percentages across all samples, so that we can pick top species
        self.metaphlan_total_pct = dict()
        self.sum_sample_counts()

        self.general_stats_cols()
        self.top_taxa_barplot()

    def parse_logs(self, f):
        """
        Parses a metaphlan output .txt file to extract the following information

        1. A rank code, indicating:
            * (k)ingdom
            * (p)hylum
            * (c)lass
            * (o)rder
            * (f)amily
            * (g)enus
            * (s)pecies
        2. Clade name
        3. Relative abundance of this clade
        """

        # Search regexes for stats
        regex_kingdom_level = re.compile(r"([kpcofgs])\_\_(\w+)\t(\d+)\t(\d{1,3}\.\d{1,5})")
        regex_rel_abundance = re.compile(r"\t(\d{1,3}\.\d{1,5})")
        regex_last_taxid = re.compile(r"\|(\d+)\t")
        regex_last_tax_rank = re.compile(r"\|([kpcofgs])\_\_(\w+)\t")
        regex_tax_unclass = re.compile(r"\|([kpcofgs])\_\_(\w+)\_(unclassified)\t")
        regex_comments = re.compile(r"\#")
        data = []
        for line in f["f"]:
            match = regex_last_tax_rank.search(line)
            match_last_id = regex_last_taxid.search(line)
            match_per = regex_rel_abundance.search(line)
            match_first = regex_kingdom_level.search(line)
            match_unclassified = regex_tax_unclass.search(line)
            match_comments = regex_comments.search(line)
            if match_first:
                # Matches the first row (kingdom)
                row = {
                    "tax_rank": match_first.group(1),
                    "taxonomy": match_first.group(2),
                    "rel_abundance": float(match_first.group(4)),
                }
                data.append(row)
            elif match_unclassified and match_per:
                # Include unclassified reads on each rank
                row = {
                    "tax_rank": match_unclassified.group(1),
                    "taxonomy": match_unclassified.group(3),
                    "rel_abundance": float(match_per.group(1)),
                }
                data.append(row)
            elif match and match_last_id and match_per:
                # Match all other clades
                row = {
                    "tax_rank": match.group(1),
                    "taxonomy": match.group(2),
                    "rel_abundance": float(match_per.group(1)),
                }
                data.append(row)
            elif match_comments:
                continue
            else:
                log.debug(f"{f['s_name']}: Could not parse line: {line}")
        self.metaphlan_raw_data[f["s_name"]] = data

    def sum_sample_counts(self):
        """Sum relative abundance across all samples for metaphlan data"""

        # Sum the percentages for each taxa across all samples
        # Allows us to pick the top taxa for each rank
        for s_name, data in self.metaphlan_raw_data.items():
            for row in data:
                tax_rank = row["tax_rank"]
                taxonomy = row["taxonomy"]

                if tax_rank not in self.metaphlan_total_pct:
                    self.metaphlan_total_pct[tax_rank] = dict()

                if taxonomy not in self.metaphlan_total_pct[tax_rank]:
                    self.metaphlan_total_pct[tax_rank][taxonomy] = 0
                self.metaphlan_total_pct[tax_rank][taxonomy] += row["rel_abundance"]

    def general_stats_cols(self):
        """Add a couple of columns to the General Statistics table"""

        # Get top taxa in most specific taxa rank that we have
        top_taxa = []
        top_rank_code = None
        top_rank_name = None
        for rank_code, rank_name in self.t_ranks.items():
            try:
                sorted_pct = sorted(
                    self.metaphlan_total_pct[rank_code].items(),
                    key=lambda x: x[1],
                    reverse=True,
                )
                for taxonomy, pct_sum in sorted_pct[: self.top_n]:
                    top_taxa.append(taxonomy)
                top_rank_code = rank_code
                top_rank_name = rank_name
                break
            except KeyError:
                # No species-level data found etc
                pass

        # Column headers
        headers = dict()

        top_one_hkey = None

        top_one_hkey = "% {}".format(top_taxa[0])
        headers[top_one_hkey] = {
            "title": top_one_hkey,
            "description": "Percentage of reads that were the top {} over all samples - {}".format(
                top_rank_name, top_taxa[0]
            ),
            "suffix": "%",
            "max": 100,
            "scale": "PuBuGn",
        }
        headers["% Top"] = {
            "title": f"% Top {self.top_n} {top_rank_name}",
            "description": f"Percentage of reads that were classified by one of the top-{self.top_n} {top_rank_name} ({', '.join(top_taxa)})",
            "suffix": "%",
            "max": 100,
            "scale": "PuBu",
        }
        # Get table data
        tdata = {}
        for s_name, d in self.metaphlan_raw_data.items():
            tdata[s_name] = {}
            for row in d:
                percent = row["rel_abundance"]
                if row["tax_rank"] == top_rank_code and row["taxonomy"] in top_taxa:
                    tdata[s_name]["% Top"] = percent + tdata[s_name].get("% Top", 0)
                if row["tax_rank"] == top_rank_code and row["taxonomy"] == top_taxa[0]:
                    tdata[s_name][top_one_hkey] = percent

            if top_one_hkey is not None and top_one_hkey not in tdata[s_name]:
                tdata[s_name][top_one_hkey] = 0

        self.general_stats_addcols(tdata, headers)

    def top_taxa_barplot(self):
        """Add a bar plot showing the top-N from each taxa rank"""

        pd = []
        cats = list()
        # Keeping track of encountered codes to display only tabs with available data
        found_rank_codes = set()

        for rank_code in self.t_ranks:
            rank_cats = dict()
            rank_data = dict()

            # Loop through the summed tax percentages to get the top-N across all samples
            try:
                sorted_pct = sorted(
                    self.metaphlan_total_pct[rank_code].items(),
                    key=lambda x: x[1],
                    reverse=True,
                )
            except KeyError:
                # Taxa rank not found in this sample
                continue
            i = 0
            counts_shown = {}
            for taxonomy, pct_sum in sorted_pct:
                # Adding unclassified for each rank:
                if taxonomy == "unclassified":
                    rank_cats[taxonomy] = {"name": taxonomy}
                    for s_name, d in self.metaphlan_raw_data.items():
                        if s_name not in rank_data:
                            rank_data[s_name] = dict()
                        if s_name not in counts_shown:
                            counts_shown[s_name] = 0
                        for row in d:
                            if row["tax_rank"] == rank_code:
                                found_rank_codes.add(rank_code)
                                if row["taxonomy"] == taxonomy:
                                    if taxonomy not in rank_data[s_name]:
                                        rank_data[s_name][taxonomy] = 0
                                    rank_data[s_name][taxonomy] += row["rel_abundance"]
                                    counts_shown[s_name] += row["rel_abundance"]
                                    rank_data[s_name]["other"] = 100 - counts_shown[s_name]
                else:
                    # Add top n taxa for each rank
                    i += 1
                    if i > self.top_n:
                        # After top 5, keep looping to sum up unclassified
                        continue
                    rank_cats[taxonomy] = {"name": taxonomy}
                    # Pull out counts for this rank + classif from each sample
                    for s_name, d in self.metaphlan_raw_data.items():
                        if s_name not in rank_data:
                            rank_data[s_name] = dict()
                        if s_name not in counts_shown:
                            counts_shown[s_name] = 0

                        for row in d:
                            if row["tax_rank"] == rank_code:
                                found_rank_codes.add(rank_code)
                                if row["taxonomy"] == taxonomy:
                                    if taxonomy not in rank_data[s_name]:
                                        rank_data[s_name][taxonomy] = 0
                                    rank_data[s_name][taxonomy] += row["rel_abundance"]
                                    counts_shown[s_name] += row["rel_abundance"]
                                    rank_data[s_name]["other"] = 100 - counts_shown[s_name]
            # Add in unclassified reads and "other" - we presume from other species etc.
            for s_name, d in self.metaphlan_raw_data.items():
                # In case none of the top_n were in some sample:
                if s_name not in rank_data:
                    rank_data[s_name] = dict()
                if s_name not in counts_shown:
                    counts_shown[s_name] = 0
                rank_data[s_name]["other"] = 100 - counts_shown[s_name]

                # This should never happen... But it does sometimes if the total read count is a bit off
                if rank_data[s_name]["other"] < 0:
                    log.debug(
                        "Found negative 'other' count for {} ({}): {}".format(
                            s_name, self.t_ranks[rank_code], rank_data[s_name]["other"]
                        )
                    )
                    rank_data[s_name]["other"] = 0

            rank_cats["other"] = {"name": "Other", "color": "#cccccc"}
            rank_cats["unclassified"] = {"name": "Unclassified", "color": "#d4949c"}

            cats.append(rank_cats)
            pd.append(rank_data)

        pconfig = {
            "id": f"{self.anchor}-top-n-plot",
            "title": f"{self.name}: Top taxa",
            "ylab": "Number of fragments",
            "data_labels": [v for k, v in self.t_ranks.items() if k in found_rank_codes],
            "cpswitch": False,
        }

        self.add_section(
            name="Top taxa",
            anchor=f"{self.anchor}-top-n",
            description=f"The relative abundance of reads falling into the top {self.top_n} taxa across different ranks.",
            helptext=f"""
                To make this plot, the percentage of each sample assigned to a given taxa is summed across all samples.
                The counts for these top {self.top_n} taxa are then plotted for each of the 7 different taxa ranks.
                The unclassified percentages are shown per taxa rank.

                The category _"Other"_ shows the difference between 100% and the sum of the percent
                in the top {self.top_n} taxa shown + the percent that was unclassified. This should cover all taxa _not_ in the top {self.top_n}, +/- any rounding errors.

            """,
            plot=bargraph.plot(pd, cats, pconfig),
        )



"""MultiQC module to plot output from goleft indexcov

https://github.com/brentp/goleft/tree/master/indexcov
"""


import collections
import logging

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import linegraph, scatter

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        super(MultiqcModule, self).__init__(
            name="goleft indexcov",
            anchor="goleft_indexcov",
            href="https://github.com/brentp/goleft/tree/master/indexcov",
            info="quickly estimates coverage from a whole-genome bam index.",
            doi="10.1093/gigascience/gix090",
        )

        # Parse ROC data
        self.roc_plot_data = collections.defaultdict(lambda: collections.defaultdict(dict))
        for f in self.find_log_files("goleft_indexcov/roc", filehandles=True):
            self.parse_roc_plot_data(f)
            self.add_data_source(f)

        # Filter to strip out ignored sample names
        num_roc_samples = 0
        for chrom in self.roc_plot_data:
            self.roc_plot_data[chrom] = self.ignore_samples(self.roc_plot_data[chrom])
            num_roc_samples = max(len(self.roc_plot_data[chrom]), num_roc_samples)

        # Write data to file
        self.write_data_file(self.roc_plot_data, "goleft_roc")

        # Parse BIN data
        self.bin_plot_data = {}
        self.bin_plot_data_empty_samples = []
        for f in self.find_log_files("goleft_indexcov/ped", filehandles=True):
            self.parse_bin_plot_data(f)
            self.add_data_source(f)

        # Filter to strip out ignored sample names
        self.bin_plot_data = self.ignore_samples(self.bin_plot_data)

        # Write data to file
        self.write_data_file(self.roc_plot_data, "goleft_bin")

        # Stop execution if no samples
        num_samples = max(len(self.bin_plot_data), num_roc_samples)
        if num_samples == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {num_samples} samples")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Add sections to the report
        self.roc_plot()
        self.bin_plot()

    def _short_chrom(self, chrom):
        """Plot standard chromosomes + X, sorted numerically.

        Allows specification from a list of chromosomes via config
        for non-standard genomes.
        """
        default_allowed = {"X"}
        allowed_chroms = set(getattr(config, "goleft_indexcov_config", {}).get("chromosomes", []))

        chrom_clean = chrom.replace("chr", "")
        try:
            chrom_clean = int(chrom_clean)
        except ValueError:
            if chrom_clean not in default_allowed and chrom_clean not in allowed_chroms:
                chrom_clean = None

        if allowed_chroms:
            if chrom in allowed_chroms or chrom_clean in allowed_chroms:
                return chrom_clean
        elif isinstance(chrom_clean, int) or chrom_clean in default_allowed:
            return chrom_clean

    def parse_roc_plot_data(self, f):
        header = f["f"].readline()
        sample_names = [self.clean_s_name(x, f) for x in header.strip().split()[2:]]
        for parts in (line.rstrip().split() for line in f["f"]):
            if len(parts) > 2:
                chrom, cov = parts[:2]
                sample_vals = parts[2:]
                if self._short_chrom(chrom) is not None:
                    for val, sample in zip(sample_vals, sample_names):
                        if not self.is_ignore_sample(sample):
                            self.roc_plot_data[chrom][sample][float(cov)] = float(val)

    def roc_plot(self):
        def to_padded_str(x):
            x = self._short_chrom(x)
            try:
                return "%06d" % x
            except TypeError:
                return x

        chroms = sorted(self.roc_plot_data.keys(), key=to_padded_str)

        max_chroms = getattr(config, "goleft_indexcov_config", {}).get("max_chroms", 50)
        if len(chroms) > max_chroms:
            log.warning(f"Too many chromosomes found: {len(chroms)}, limiting to {max_chroms}")
            chroms = chroms[:max_chroms]

        pconfig = {
            "id": "goleft_indexcov-roc-plot",
            "title": "goleft indexcov: ROC - genome coverage per scaled depth by chromosome",
            "xlab": "Scaled coverage",
            "ylab": "Proportion of regions covered",
            "ymin": 0,
            "ymax": 1.0,
            "xmin": 0,
            "xmax": 1.5,
            "data_labels": [{"name": self._short_chrom(c)} for c in chroms],
        }
        self.add_section(
            name="Scaled coverage ROC plot",
            anchor="goleft_indexcov-roc",
            description="Coverage (ROC) plot that shows genome coverage at given (scaled) depth.",
            helptext="""
                Lower coverage samples have shorter curves where the proportion of regions covered
                drops off more quickly. This indicates a higher fraction of low coverage regions.
            """,
            plot=linegraph.plot([self.roc_plot_data[c] for c in chroms], pconfig),
        )

    def parse_bin_plot_data(self, f):
        header = f["f"].readline()[1:].strip().split("\t")
        for sample_parts in (line.split("\t") for line in f["f"]):
            cur = dict(zip(header, sample_parts))
            cur["sample_id"] = self.clean_s_name(cur["sample_id"], f)
            total = float(cur["bins.in"]) + float(cur["bins.out"])
            try:
                self.bin_plot_data[cur["sample_id"]] = {
                    "x": float(cur["bins.lo"]) / total,
                    "y": float(cur["bins.out"]) / total,
                }
            except ZeroDivisionError:
                self.bin_plot_data_empty_samples.append(cur["sample_id"])
        return self.bin_plot_data_empty_samples

    def bin_plot(self):
        pconfig = {
            "id": "goleft_indexcov-bin-plot",
            "title": "goleft indexcov: Problematic low and non-uniform coverage bins",
            "xlab": "Proportion of bins with depth < 0.15",
            "ylab": "Proportion of bins with depth outside of (0.85, 1.15)",
            "yCeiling": 1.0,
            "yFloor": 0.0,
            "xCeiling": 1.0,
            "xFloor": 0.0,
        }
        extra = ""
        if len(self.bin_plot_data_empty_samples) > 0:
            # Bootstrap alert about missing samples
            extra = f"""<div class="alert alert-warning" style="margin:2rem 0;">
                <strong>Warning:</strong>
                {len(self.bin_plot_data_empty_samples)} sample{'s' if len(self.bin_plot_data_empty_samples) > 1 else ''} had zero bins and could not be plotted.
                <a href="#goleft_empty_samples" onclick="$('#goleft_empty_samples').slideToggle();">Click to show missing sample names.</a>
                <div id="goleft_empty_samples" style="display:none;">
                    <ul><li><code>{"</code></li>, <li><code>".join(self.bin_plot_data_empty_samples)}</code></li></ul>
                </div>
            </div>"""
        self.add_section(
            name="Problem coverage bins",
            anchor="goleft_indexcov-bin",
            description="This plot identifies problematic samples using binned coverage distributions.",
            helptext="""
                We expect bins to be around 1, so deviations from this indicate problems.
                Low coverage bins (`< 0.15`) on the x-axis have regions with low or missing coverage.
                Higher values indicate truncated BAM files or missing data.
                Bins with skewed distributions (`<0.85` or `>1.15`) on the y-axis detect dosage bias.
                Large values on the y-axis are likely to impact CNV and structural variant calling.
                See the [goleft indexcov bin documentation](https://github.com/brentp/goleft/blob/master/docs/indexcov/help-bin.md)
                for more details.
            """,
            plot=scatter.plot(self.bin_plot_data, pconfig),
            content=extra,
        )
